import collections

from ...core import Expr, Integer
from ...core.sympify import sympify
from ...logic import true
from ...matrices import MatrixBase
from ...printing.defaults import DefaultPrinting
from ..indexed import Indexed


class NDimArray(DefaultPrinting):
    """N-dim array.

    Examples
    ========

    Create an N-dim array of zeros:

    >>> a = MutableDenseNDimArray.zeros(2, 3, 4)
    >>> a
    [[[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
     [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]]

    Create an N-dim array from a list;

    >>> a = MutableDenseNDimArray([[2, 3], [4, 5]])
    >>> a
    [[2, 3], [4, 5]]

    >>> b = MutableDenseNDimArray([[[1, 2], [3, 4], [5, 6]],
    ...                            [[7, 8], [9, 10], [11, 12]]])
    >>> b
    [[[1, 2], [3, 4], [5, 6]], [[7, 8], [9, 10], [11, 12]]]

    Create an N-dim array from a flat list with dimension shape:

    >>> a = MutableDenseNDimArray([1, 2, 3, 4, 5, 6], (2, 3))
    >>> a
    [[1, 2, 3], [4, 5, 6]]

    Create an N-dim array from a matrix:

    >>> a = Matrix([[1, 2], [3, 4]])
    >>> a
    Matrix([
    [1, 2],
    [3, 4]])
    >>> b = MutableDenseNDimArray(a)
    >>> b
    [[1, 2], [3, 4]]

    Arithmetic operations on N-dim arrays

    >>> a = MutableDenseNDimArray([1, 1, 1, 1], (2, 2))
    >>> b = MutableDenseNDimArray([4, 4, 4, 4], (2, 2))
    >>> c = a + b
    >>> c
    [[5, 5], [5, 5]]
    >>> a - b
    [[-3, -3], [-3, -3]]

    """

    def _parse_index(self, index):

        if isinstance(index, (int, Integer)):
            if index >= self._loop_size:
                raise ValueError('index out of range')
            return index

        if len(index) != self._rank:
            raise ValueError('Wrong number of array axes')

        real_index = 0
        # check if input index can exist in current indexing
        for i in range(self._rank):
            if index[i] >= self.shape[i]:
                raise ValueError('Index ' + str(index) + ' out of border')
            real_index = real_index*self.shape[i] + index[i]

        return real_index

    def _get_tuple_index(self, integer_index):
        index = []
        for sh in reversed(self.shape):
            index.append(integer_index % sh)
            integer_index //= sh
        index.reverse()
        return tuple(index)

    def _check_symbolic_index(self, index):
        # Check if any index is symbolic:
        tuple_index = (index if isinstance(index, tuple) else (index,))
        if any((isinstance(i, Expr) and (not i.is_number)) for i in tuple_index):
            for i, nth_dim in zip(tuple_index, self.shape):
                i = sympify(i)
                if ((i < 0) is true) or ((i >= nth_dim) is true):
                    raise ValueError('index out of range')
            return Indexed(self, *tuple_index)

    def _setter_iterable_check(self, value):
        if isinstance(value, (collections.abc.Iterable, MatrixBase, NDimArray)):
            raise NotImplementedError

    @classmethod
    def _scan_iterable_shape(cls, iterable):
        def f(pointer):
            if not isinstance(pointer, collections.abc.Iterable):
                return [pointer], ()

            result = []
            elems, shapes = zip(*[f(i) for i in pointer])
            if len(set(shapes)) != 1:
                raise ValueError('could not determine shape unambiguously')
            for i in elems:
                result.extend(i)
            return result, (len(shapes),)+shapes[0]

        return f(iterable)

    @classmethod
    def _handle_ndarray_creation_inputs(cls, iterable=None, shape=None, **kwargs):

        if shape is None and iterable is None:
            shape = ()
            iterable = ()

        # Construction from another `NDimArray`:
        elif shape is None and isinstance(iterable, NDimArray):
            shape = iterable.shape
            iterable = list(iterable)

        # Construct N-dim array from an iterable (numpy arrays included):
        elif shape is None and isinstance(iterable, collections.abc.Iterable):
            iterable, shape = cls._scan_iterable_shape(iterable)

        # Construct N-dim array from a Matrix:
        elif shape is None and isinstance(iterable, MatrixBase):
            shape = iterable.shape

        # Construct NDimArray(iterable, shape)
        elif shape is not None:
            pass

        else:
            raise TypeError('Data type not understood')

        if isinstance(shape, (int, Integer)):
            shape = shape,

        shape = tuple(shape)

        if any(not isinstance(dim, (int, Integer)) for dim in shape):
            raise TypeError('Shape should contain integers only.')

        if isinstance(iterable, collections.abc.Mapping):
            for k in list(iterable):
                if not isinstance(k, collections.abc.Sequence):
                    continue
                new_key = 0
                for i, idx in enumerate(k):
                    new_key = new_key * shape[i] + idx
                iterable[new_key] = iterable[k]
                del iterable[k]

        return shape, iterable

    def __len__(self):
        """Overload common function len(). Returns number of elements in array.

        Examples
        ========

        >>> a = MutableDenseNDimArray.zeros(3, 3)
        >>> a
        [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        >>> len(a)
        9

        """
        return self._loop_size

    @property
    def shape(self):
        """
        Returns array shape (dimension).

        Examples
        ========

        >>> a = MutableDenseNDimArray.zeros(3, 3)
        >>> a.shape
        (3, 3)

        """
        return self._shape

    def rank(self):
        """
        Returns rank of array.

        Examples
        ========

        >>> a = MutableDenseNDimArray.zeros(3, 4, 5, 6, 3)
        >>> a.rank()
        5

        """
        return self._rank

    def diff(self, *args, **kwargs):
        """
        Calculate the derivative of each element in the array.

        Examples
        ========

        >>> M = ImmutableDenseNDimArray([[x, y], [1, x*y]])
        >>> M.diff(x)
        [[1, 0], [0, y]]

        """
        # pylint: disable=not-an-iterable
        return type(self)((x.diff(*args, **kwargs) for x in self), self.shape)

    def applyfunc(self, f):
        """Apply a function to each element of the N-dim array.

        Examples
        ========

        >>> m = ImmutableDenseNDimArray([i*2+j for i in range(2)
        ...                              for j in range(2)], (2, 2))
        >>> m
        [[0, 1], [2, 3]]
        >>> m.applyfunc(lambda i: 2*i)
        [[0, 2], [4, 6]]

        """
        return type(self)(map(f, self), self.shape)

    def __str__(self):
        """Returns string, allows to use standard functions print() and str().

        Examples
        ========

        >>> a = MutableDenseNDimArray.zeros(2, 2)
        >>> a
        [[0, 0], [0, 0]]

        """
        def f(sh, shape_left, i, j):
            # pylint: disable=unsubscriptable-object
            if len(shape_left) == 1:
                return '['+', '.join([str(self[e]) for e in range(i, j)])+']'

            sh //= shape_left[0]
            return '[' + ', '.join([f(sh, shape_left[1:], i+e*sh, i+(e+1)*sh) for e in range(shape_left[0])]) + ']'  # + '\n'*len(shape_left)

        return f(self._loop_size, self.shape, 0, self._loop_size)

    def tolist(self):
        """
        Conveting MutableDenseNDimArray to one-dim list

        Examples
        ========

        >>> a = MutableDenseNDimArray([1, 2, 3, 4], (2, 2))
        >>> a
        [[1, 2], [3, 4]]
        >>> b = a.tolist()
        >>> b
        [[1, 2], [3, 4]]

        """

        def f(sh, shape_left, i, j):
            # pylint: disable=unsubscriptable-object
            if len(shape_left) == 1:
                return [self[e] for e in range(i, j)]
            result = []
            sh //= shape_left[0]
            for e in range(shape_left[0]):
                result.append(f(sh, shape_left[1:], i+e*sh, i+(e+1)*sh))
            return result

        return f(self._loop_size, self.shape, 0, self._loop_size)

    def __add__(self, other):
        if not isinstance(other, NDimArray):
            raise TypeError(str(other))

        if self.shape != other.shape:
            raise ValueError('array shape mismatch')
        result_list = [i + j for i, j in zip(self, other)]

        return type(self)(result_list, self.shape)

    def __sub__(self, other):
        if not isinstance(other, NDimArray):
            raise TypeError(str(other))

        if self.shape != other.shape:
            raise ValueError('array shape mismatch')
        result_list = [i - j for i, j in zip(self, other)]

        return type(self)(result_list, self.shape)

    def __mul__(self, other):
        if isinstance(other, (collections.abc.Iterable, NDimArray, MatrixBase)):
            raise ValueError('scalar expected, use tensorproduct(...) for tensorial product')
        other = sympify(other)
        result_list = [i*other for i in self]  # pylint: disable=not-an-iterable
        return type(self)(result_list, self.shape)

    def __rmul__(self, other):
        if isinstance(other, (collections.abc.Iterable, NDimArray, MatrixBase)):
            raise ValueError('scalar expected, use tensorproduct(...) for tensorial product')
        other = sympify(other)
        result_list = [other*i for i in self]  # pylint: disable=not-an-iterable
        return type(self)(result_list, self.shape)

    def __truediv__(self, other):
        if isinstance(other, (collections.abc.Iterable, NDimArray, MatrixBase)):
            raise ValueError('scalar expected')
        other = sympify(other)
        result_list = [i/other for i in self]  # pylint: disable=not-an-iterable
        return type(self)(result_list, self.shape)

    def __rtruediv__(self, other):
        return NotImplemented

    def __eq__(self, other):
        """
        Compare NDimArray instances.

        Instances equal if they have same shape and data.

        Examples
        ========

        >>> a = MutableDenseNDimArray.zeros(2, 3)
        >>> b = MutableDenseNDimArray.zeros(2, 3)
        >>> a == b
        True
        >>> c = a.reshape(3, 2)
        >>> c == b
        False
        >>> a[0, 0] = 1
        >>> b[0, 0] = 2
        >>> a == b
        False

        """
        if not isinstance(other, NDimArray):
            return False
        return (self.shape == other.shape) and (list(self) == list(other))

    def _eval_transpose(self):
        from .arrayop import permutedims
        if self.rank() != 2:
            raise ValueError('array rank not 2')
        return permutedims(self, (1, 0))

    def transpose(self):
        return self._eval_transpose()

    def _eval_conjugate(self):
        return self.func([i.conjugate() for i in self], self.shape)  # pylint: disable=not-an-iterable

    def conjugate(self):
        return self._eval_conjugate()

    def _eval_adjoint(self):
        return self.transpose().conjugate()

    def adjoint(self):
        return self._eval_adjoint()


class ImmutableNDimArray(NDimArray, Expr):
    """An immutable version of the N-dim array."""

    _op_priority = 11.0



"""Tests for useful utilities for higher level polynomial classes."""

import pytest

from diofant import (ZZ, GeneratorsNeeded, I, Integer, Integral, Mul,
                     PolynomialError, Rational, cos, erf, exp, factor,
                     integrate, pi, sin, sqrt, symbols)
from diofant.abc import p, q, t, x, y, z
from diofant.polys.polyutils import (_nsort, _sort_factors, _sort_gens,
                                     _unify_gens, parallel_dict_from_expr)


__all__ = ()

A, B = symbols('A,B', commutative=False)


def test__nsort():
    # issue sympy/sympy#6137
    r = ([Rational(3, 2) + sqrt(Rational(-14, 3) - 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3) -
                                4/sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) +
                                       2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) -
                                61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)))/2 -
          sqrt(-7/3 + 61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) +
               2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3))/2,
          Rational(3, 2) - sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) +
                                2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3))/2 -
          sqrt(Rational(-14, 3) - 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3) -
               4/sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) +
                      2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) -
               61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)))/2, Rational(3, 2) +
          sqrt(Rational(-14, 3) - 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3) +
               4/sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) +
                                                13*I/12)**Rational(1, 3)) + 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) -
               61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)))/2 +
          sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216)
                                         + 13*I/12)**Rational(1, 3)) + 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3))/2,
          Rational(3, 2) + sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) +
                                                          13*I/12)**Rational(1, 3)) + 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3))/2 -
          sqrt(Rational(-14, 3) - 2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3) +
               4/sqrt(Rational(-7, 3) + 61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) +
                      2*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)) -
               61/(18*(Rational(-415, 216) + 13*I/12)**Rational(1, 3)))/2])
    ans = [r[1], r[0], r[-1], r[-2]]
    assert _nsort(r) == ans
    assert len(_nsort(r, separated=True)[0]) == 0
    b, c, a = exp(-1000), exp(-999), exp(-1001)
    assert _nsort((b, c, a)) == [a, b, c]
    d = symbols('d', extended_real=True)
    assert _nsort((d,)) == [d]
    assert _nsort((d,), separated=True) == [[d], []]
    c = symbols('c', complex=True, real=False)
    assert _nsort((c,)) == [c]
    assert _nsort((c,), separated=True) == [[], [c]]
    assert _nsort((I, Rational(1)), separated=True) == ([Rational(1)], [I])


def test__sort_gens():
    assert not _sort_gens([])

    assert _sort_gens([x]) == (x,)
    assert _sort_gens([p]) == (p,)
    assert _sort_gens([q]) == (q,)

    assert _sort_gens([x, p]) == (x, p)
    assert _sort_gens([p, x]) == (x, p)
    assert _sort_gens([q, p]) == (p, q)

    assert _sort_gens([q, p, x]) == (x, p, q)

    assert _sort_gens([x, p, q], wrt=x) == (x, p, q)
    assert _sort_gens([x, p, q], wrt=p) == (p, x, q)
    assert _sort_gens([x, p, q], wrt=q) == (q, x, p)

    assert _sort_gens([x, p, q], wrt='x') == (x, p, q)
    assert _sort_gens([x, p, q], wrt='p') == (p, x, q)
    assert _sort_gens([x, p, q], wrt='q') == (q, x, p)

    assert _sort_gens([x, p, q], wrt='x,q') == (x, q, p)
    assert _sort_gens([x, p, q], wrt='q,x') == (q, x, p)
    assert _sort_gens([x, p, q], wrt='p,q') == (p, q, x)
    assert _sort_gens([x, p, q], wrt='q,p') == (q, p, x)

    assert _sort_gens([x, p, q], wrt='x, q') == (x, q, p)
    assert _sort_gens([x, p, q], wrt='q, x') == (q, x, p)
    assert _sort_gens([x, p, q], wrt='p, q') == (p, q, x)
    assert _sort_gens([x, p, q], wrt='q, p') == (q, p, x)

    assert _sort_gens([x, p, q], wrt=[x, 'q']) == (x, q, p)
    assert _sort_gens([x, p, q], wrt=[q, 'x']) == (q, x, p)
    assert _sort_gens([x, p, q], wrt=[p, 'q']) == (p, q, x)
    assert _sort_gens([x, p, q], wrt=[q, 'p']) == (q, p, x)

    assert _sort_gens([x, p, q], wrt=['x', 'q']) == (x, q, p)
    assert _sort_gens([x, p, q], wrt=['q', 'x']) == (q, x, p)
    assert _sort_gens([x, p, q], wrt=['p', 'q']) == (p, q, x)
    assert _sort_gens([x, p, q], wrt=['q', 'p']) == (q, p, x)

    assert _sort_gens([x, p, q], sort='x > p > q') == (x, p, q)
    assert _sort_gens([x, p, q], sort='p > x > q') == (p, x, q)
    assert _sort_gens([x, p, q], sort='p > q > x') == (p, q, x)

    assert _sort_gens([x, p, q], wrt='x', sort='q > p') == (x, q, p)
    assert _sort_gens([x, p, q], wrt='p', sort='q > x') == (p, q, x)
    assert _sort_gens([x, p, q], wrt='q', sort='p > x') == (q, p, x)

    X = symbols('x0:3 x10:13 x20:23')

    assert _sort_gens(X) == X


def test__unify_gens():
    assert not _unify_gens([], [])

    assert _unify_gens([x], [x]) == (x,)
    assert _unify_gens([y], [y]) == (y,)

    assert _unify_gens([x, y], [x]) == (x, y)
    assert _unify_gens([x], [x, y]) == (x, y)

    assert _unify_gens([x, y], [x, y]) == (x, y)
    assert _unify_gens([y, x], [y, x]) == (y, x)

    assert _unify_gens([x], [y]) == (x, y)
    assert _unify_gens([y], [x]) == (y, x)

    assert _unify_gens([x], [y, x]) == (y, x)
    assert _unify_gens([y, x], [x]) == (y, x)

    assert _unify_gens([x, y, z], [x, y, z]) == (x, y, z)
    assert _unify_gens([z, y, x], [x, y, z]) == (z, y, x)
    assert _unify_gens([x, y, z], [z, y, x]) == (x, y, z)
    assert _unify_gens([z, y, x], [z, y, x]) == (z, y, x)

    assert _unify_gens([x, y, z], [t, x, p, q, z]) == (t, x, y, p, q, z)


def test__sort_factors():
    assert _sort_factors([], multiple=True) == []
    assert _sort_factors([], multiple=False) == []

    F = [[1, 2, 3], [1, 2], [1]]
    G = [[1], [1, 2], [1, 2, 3]]

    assert _sort_factors(F, multiple=False) == G

    F = [[1, 2], [1, 2, 3], [1, 2], [1]]
    G = [[1], [1, 2], [1, 2], [1, 2, 3]]

    assert _sort_factors(F, multiple=False) == G

    F = [[2, 2], [1, 2, 3], [1, 2], [1]]
    G = [[1], [1, 2], [2, 2], [1, 2, 3]]

    assert _sort_factors(F, multiple=False) == G

    F = [([1, 2, 3], 1), ([1, 2], 1), ([1], 1)]
    G = [([1], 1), ([1, 2], 1), ([1, 2, 3], 1)]

    assert _sort_factors(F, multiple=True) == G

    F = [([1, 2], 1), ([1, 2, 3], 1), ([1, 2], 1), ([1], 1)]
    G = [([1], 1), ([1, 2], 1), ([1, 2], 1), ([1, 2, 3], 1)]

    assert _sort_factors(F, multiple=True) == G

    F = [([2, 2], 1), ([1, 2, 3], 1), ([1, 2], 1), ([1], 1)]
    G = [([1], 1), ([1, 2], 1), ([2, 2], 1), ([1, 2, 3], 1)]

    assert _sort_factors(F, multiple=True) == G

    F = [([2, 2], 1), ([1, 2, 3], 1), ([1, 2], 2), ([1], 1)]
    G = [([1], 1), ([2, 2], 1), ([1, 2], 2), ([1, 2, 3], 1)]

    assert _sort_factors(F, multiple=True) == G


def test__dict_from_expr_if_gens():
    assert parallel_dict_from_expr([Integer(17)],
                                   gens=(x,)) == ([{(0,): 17}], (x,))
    assert parallel_dict_from_expr([Integer(17)],
                                   gens=(x, y)) == ([{(0, 0): 17}], (x, y))
    assert parallel_dict_from_expr([Integer(17)],
                                   gens=(x, y, z)) == ([{(0, 0, 0): 17}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([Integer(-17)],
                                   gens=(x,)) == ([{(0,): -17}], (x,))
    assert parallel_dict_from_expr([Integer(-17)],
                                   gens=(x, y)) == ([{(0, 0): -17}], (x, y))
    assert parallel_dict_from_expr([Integer(-17)],
                                   gens=(x, y, z)) == ([{(0, 0, 0): -17}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([17*x], gens=(x,)) == ([{(1,): 17}], (x,))
    assert parallel_dict_from_expr([17*x],
                                   gens=(x, y)) == ([{(1, 0): 17}], (x, y))
    assert parallel_dict_from_expr([17*x],
                                   gens=(x, y, z)) == ([{(1, 0, 0): 17}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([17*x**7], gens=(x,)) == ([{(7,): 17}], (x,))
    assert parallel_dict_from_expr([17*x**7*y],
                                   gens=(x, y)) == ([{(7, 1): 17}], (x, y))
    assert parallel_dict_from_expr([17*x**7*y*z**12],
                                   gens=(x, y, z)) == ([{(7, 1, 12): 17}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([x + 2*y + 3*z],
                                   gens=(x,)) == ([{(1,): 1,
                                                    (0,): 2*y + 3*z}], (x,))
    assert parallel_dict_from_expr([x + 2*y + 3*z],
                                   gens=(x, y)) == ([{(1, 0): 1, (0, 1): 2,
                                                      (0, 0): 3*z}], (x, y))
    assert parallel_dict_from_expr([x + 2*y + 3*z],
                                   gens=(x, y, z)) == ([{(1, 0, 0): 1,
                                                         (0, 1, 0): 2,
                                                         (0, 0, 1): 3}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([x*y + 2*x*z + 3*y*z],
                                   gens=(x,)) == ([{(1,): y + 2*z,
                                                    (0,): 3*y*z}], (x,))
    assert parallel_dict_from_expr([x*y + 2*x*z + 3*y*z],
                                   gens=(x, y)) == ([{(1, 1): 1, (1, 0): 2*z,
                                                      (0, 1): 3*z}], (x, y))
    assert parallel_dict_from_expr([x*y + 2*x*z + 3*y*z],
                                   gens=(x, y, z)) == ([{(1, 1, 0): 1,
                                                         (1, 0, 1): 2,
                                                         (0, 1, 1): 3}],
                                                       (x, y, z))

    assert parallel_dict_from_expr([2**y*x],
                                   gens=(x,)) == ([{(1,): 2**y}], (x,))
    assert parallel_dict_from_expr([Integral(x, (x, 1, 2)) +
                                    x]) == ([{(0, 1): 1, (1, 0): 1}],
                                            (x, Integral(x, (x, 1, 2))))

    pytest.raises(PolynomialError,
                  lambda: parallel_dict_from_expr([2**y*x], gens=(x, y)))


def test__dict_from_expr_no_gens():
    pytest.raises(GeneratorsNeeded,
                  lambda: parallel_dict_from_expr([Integer(17)]))

    assert parallel_dict_from_expr([x]) == ([{(1,): 1}], (x,))
    assert parallel_dict_from_expr([y]) == ([{(1,): 1}], (y,))

    assert parallel_dict_from_expr([x*y]) == ([{(1, 1): 1}], (x, y))
    assert parallel_dict_from_expr([x + y]) == ([{(1, 0): 1, (0, 1): 1}],
                                                (x, y))

    assert parallel_dict_from_expr([sqrt(2)]) == ([{(1,): 1}], (sqrt(2),))
    pytest.raises(GeneratorsNeeded,
                  lambda: parallel_dict_from_expr([sqrt(2)], greedy=False))

    assert parallel_dict_from_expr([x*y],
                                   domain=ZZ.inject(x)) == ([{(1,): x}], (y,))
    assert parallel_dict_from_expr([x*y],
                                   domain=ZZ.inject(y)) == ([{(1,): y}], (x,))

    assert parallel_dict_from_expr([3*sqrt(2)*pi*x*y],
                                   extension=None) == ([{(1, 1, 1, 1): 3}],
                                                       (x, y, pi, sqrt(2)))
    assert parallel_dict_from_expr([3*sqrt(2)*pi*x*y],
                                   extension=True) == ([{(1, 1, 1): 3*sqrt(2)}],
                                                       (x, y, pi))

    f = cos(x)*sin(x) + cos(x)*sin(y) + cos(y)*sin(x) + cos(y)*sin(y)

    assert parallel_dict_from_expr([f]) == ([{(0, 1, 0, 1): 1, (0, 1, 1, 0): 1,
                                              (1, 0, 0, 1): 1,
                                              (1, 0, 1, 0): 1}],
                                            (cos(x), cos(y), sin(x), sin(y)))


def test__parallel_dict_from_expr_if_gens():
    assert parallel_dict_from_expr([x + 2*y + 3*z, Integer(7)], gens=(x,)) == \
        ([{(1,): 1, (0,): 2*y + 3*z}, {(0,): 7}], (x,))
    assert parallel_dict_from_expr((Mul(x, x**2, evaluate=False),), gens=(x,)) == \
        ([{(3,): 1}], (x,))

    pytest.raises(PolynomialError, lambda: parallel_dict_from_expr((A*x,), gens=(x,)))


def test__parallel_dict_from_expr_no_gens():
    assert parallel_dict_from_expr([x*y, Integer(3)]) == \
        ([{(1, 1): 1}, {(0, 0): 3}], (x, y))
    assert parallel_dict_from_expr([x*y, 2*z, Integer(3)]) == \
        ([{(1, 1, 0): 1}, {(0, 0, 1): 2}, {(0, 0, 0): 3}], (x, y, z))
    assert parallel_dict_from_expr((Mul(x, x**2, evaluate=False),)) == \
        ([{(3,): 1}], (x,))


def test_parallel_dict_from_expr():
    assert parallel_dict_from_expr([x - 1, x**2 - 2]) == ([{(0,): -1, (1,): 1},
                                                           {(0,): -2,
                                                            (2,): 1}], (x,))
    pytest.raises(PolynomialError, lambda: parallel_dict_from_expr([A*B - B*A]))


def test_dict_from_expr():
    assert parallel_dict_from_expr([x - 1]) == ([{(0,): -1, (1,): 1}], (x,))
    pytest.raises(PolynomialError, lambda: parallel_dict_from_expr([A*B - B*A]))


def test_sympyissue_7383():
    x, z, R, a = symbols('x z R a')
    r = sqrt(x**2 + z**2)
    u = erf(a*r/sqrt(2))/r
    Ec = u.diff(z, z).subs({x: sqrt(R*R - z*z)})
    assert integrate(Ec, (z, -R, R)).simplify() == \
        -2*sqrt(2)*R*a**3*exp(-R**2*a**2/2)/(3*sqrt(pi))


def test_sympyissue_10161():
    x = symbols('x', real=True)
    h = (2*x*(-2*x + abs(x))*(x**2 - 1)/abs(x**2 - 1)
         + (x/abs(x) - 2)*abs(x**2 - 1))
    assert (h - factor(h)).simplify() == 0



"""
Various tests on satisfiability using the DIMACS CNF file format.

References
==========

* http://archive.dimacs.rutgers.edu/pub/challenge/satisfiability/benchmarks/cnf/

"""

import os

import pytest

from diofant.logic import satisfiable
from diofant.logic.utils import parse_dimacs


__all__ = ()


def load_file(location):
    """Loads a boolean expression from a file."""
    location = os.path.dirname(__file__) + '/' + location + '.cnf'
    with open(location, encoding='utf-8') as f:
        s = f.read()

    return parse_dimacs(s)


def test_parse_dimacs():
    pytest.raises(ValueError, lambda: parse_dimacs('q 1 2 3'))
    pytest.raises(ValueError, lambda: parse_dimacs('p cnf 2 1\n1 ?'))
    pytest.raises(ValueError, lambda: parse_dimacs('p cnf 2 2\n 1'))


@pytest.mark.parametrize('algorithm', ['dpll', 'dpll2'])
@pytest.mark.parametrize('name', ['f3', 'f5', 'quinn', 'simple_v3_c2'])
def test_dimacs_satisfiable(algorithm, name):
    assert bool(satisfiable(load_file(name), algorithm=algorithm))


@pytest.mark.parametrize('algorithm', ['dpll', 'dpll2'])
@pytest.mark.parametrize('name', ['aim-50-2_0-no-2', 'hole6'])
def test_dimacs_not_satisfiable(algorithm, name):
    assert not bool(satisfiable(load_file(name), algorithm=algorithm))



import pytest

from diofant import (Adjoint, FunctionMatrix, Identity, ImmutableMatrix,
                     Lambda, MatAdd, MatMul, MatPow, Matrix, MatrixExpr,
                     MatrixSymbol, ShapeError, Sum, Trace, ZeroMatrix, adjoint,
                     conjugate, eye, symbols, trace, transpose)


__all__ = ()

n = symbols('n', integer=True)
A = MatrixSymbol('A', n, n)
B = MatrixSymbol('B', n, n)
C = MatrixSymbol('C', 3, 4)


def test_Trace():
    assert isinstance(Trace(A), Trace)
    assert not isinstance(Trace(A), MatrixExpr)
    pytest.raises(ShapeError, lambda: Trace(C))
    assert trace(eye(3)) == 3
    assert trace(Matrix(3, 3, [1, 2, 3, 4, 5, 6, 7, 8, 9])) == 15

    assert adjoint(Trace(A)) == trace(Adjoint(A))
    assert conjugate(Trace(A)) == trace(Adjoint(A))
    assert transpose(Trace(A)) == Trace(A)

    assert isinstance(A/Trace(A), MatrixExpr)

    # Some easy simplifications
    assert trace(Identity(5)) == 5
    assert trace(ZeroMatrix(5, 5)) == 0
    assert trace(2*A*B) == 2*Trace(A*B)
    assert trace(A.T) == trace(A)

    i, j = symbols('i j')
    F = FunctionMatrix(3, 3, Lambda((i, j), i + j))
    assert trace(F) == (0 + 0) + (1 + 1) + (2 + 2)

    pytest.raises(TypeError, lambda: Trace(1))

    assert Trace(A).arg is A

    assert str(trace(A)) == str(Trace(A).doit())


def test_Trace_A_plus_B():
    assert trace(A + B) == Trace(A) + Trace(B)
    assert Trace(A + B).arg == MatAdd(A, B)
    assert Trace(A + B).doit() == Trace(A) + Trace(B)


def test_Trace_MatAdd_doit():
    # See issue sympy/sympy#9028
    X = ImmutableMatrix([[1, 2, 3]]*3)
    Y = MatrixSymbol('Y', 3, 3)
    q = MatAdd(X, 2*X, Y, -3*Y)
    assert Trace(q).arg == q
    assert Trace(q).doit() == 18 - 2*Trace(Y)


def test_Trace_MatPow_doit():
    X = Matrix([[1, 2], [3, 4]])
    assert Trace(X).doit() == 5
    q = MatPow(X, 2)
    assert Trace(q).arg == q
    assert Trace(q).doit() == 29


def test_Trace_doit_deep_False():
    X = Matrix([[1, 2], [3, 4]])
    assert Trace(X).doit(deep=False) == 5
    q = MatPow(X, 2)
    assert Trace(q).doit(deep=False).arg == q
    q = MatAdd(X, 2*X)
    assert Trace(q).doit(deep=False).arg == q
    q = MatMul(X, 2*X)
    assert Trace(q).doit(deep=False).arg == q


def test_trace_constant_factor():
    # Issue sympy/sympy#9052: gave 2*Trace(MatMul(A)) instead of 2*Trace(A)
    assert trace(2*A) == 2*Trace(A)
    X = ImmutableMatrix([[1, 2], [3, 4]])
    assert trace(MatMul(2, X)) == 10


def test_rewrite():
    assert isinstance(trace(A).rewrite(Sum), Sum)



from diofant import Symbol, exp
from diofant.abc import a, b, x


__all__ = ()


def test_equal():
    # pylint: disable=unneeded-not
    e1 = a + b
    e2 = 2*a*b
    e3 = a**3*b**2
    e4 = a*b + b*a
    assert not e1 == e2
    assert not e1 == e2
    assert e1 != e2
    assert e2 == e4
    assert e2 != e3
    assert not e2 == e3

    e1 = exp(x + 1/x)
    y = Symbol('x')
    e2 = exp(y + 1/y)
    assert e1 == e2
    assert not e1 != e2
    y = Symbol('y')
    e2 = exp(y + 1/y)
    assert not e1 == e2
    assert e1 != e2

    e5 = 3 + 2*x - x - x
    assert e5 == 3
    assert 3 == e5
    assert e5 != 4
    assert 4 != e5
    assert e5 != 3 + x
    assert 3 + x != e5


def test_expevalbug():
    e1 = exp(1*x)
    e3 = exp(x)
    assert e1 == e3


def test_cmp_bug1():
    class T:
        pass

    t = T()

    assert not x == t  # pylint: disable=unneeded-not
    assert x != t


def test_cmp_bug2():
    class T:
        pass

    t = T()

    assert not Symbol == t  # pylint: disable=unneeded-not
    assert Symbol != t


def test_sympyissue_4357():
    # pylint: disable=unneeded-not
    assert not Symbol == 1
    assert Symbol != 1
    assert not Symbol == 'x'
    assert Symbol != 'x'



"""Tests for functions for generating interesting polynomials."""

import random

import pytest

from diofant import (ZZ, cyclotomic_poly, interpolating_poly, random_poly,
                     ring, swinnerton_dyer_poly, symbols, symmetric_poly)
from diofant.abc import x, y, z


__all__ = ()


def test_swinnerton_dyer_poly():
    pytest.raises(ValueError, lambda: swinnerton_dyer_poly(0, x))

    assert swinnerton_dyer_poly(1, x, polys=True) == (x**2 - 2).as_poly()
    assert swinnerton_dyer_poly(1, polys=True) == (x**2 - 2).as_poly()

    assert swinnerton_dyer_poly(1, x) == x**2 - 2
    assert swinnerton_dyer_poly(2, x) == x**4 - 10*x**2 + 1
    assert swinnerton_dyer_poly(3, x) == (x**8 - 40*x**6 +
                                          352*x**4 - 960*x**2 + 576)
    assert swinnerton_dyer_poly(4, x) == (x**16 - 136*x**14 + 6476*x**12 -
                                          141912*x**10 + 1513334*x**8 -
                                          7453176*x**6 + 13950764*x**4 -
                                          5596840*x**2 + 46225)


def test_cyclotomic_poly():
    pytest.raises(ValueError, lambda: cyclotomic_poly(0, x))

    assert cyclotomic_poly(1, x, polys=True) == (x - 1).as_poly()
    assert cyclotomic_poly(1, polys=True) == (x - 1).as_poly()

    assert cyclotomic_poly(1, x) == x - 1
    assert cyclotomic_poly(2, x) == x + 1
    assert cyclotomic_poly(3, x) == x**2 + x + 1
    assert cyclotomic_poly(4, x) == x**2 + 1
    assert cyclotomic_poly(5, x) == x**4 + x**3 + x**2 + x + 1
    assert cyclotomic_poly(6, x) == x**2 - x + 1


def test_symmetric_poly():
    pytest.raises(ValueError, lambda: symmetric_poly(-1, x, y, z))
    pytest.raises(ValueError, lambda: symmetric_poly(5, x, y, z))

    assert symmetric_poly(1, x, y, z, polys=True) == (x + y + z).as_poly()

    assert symmetric_poly(0, x, y, z) == 1
    assert symmetric_poly(1, x, y, z) == x + y + z
    assert symmetric_poly(2, x, y, z) == x*y + x*z + y*z
    assert symmetric_poly(3, x, y, z) == x*y*z


def test_random_poly():
    poly = random_poly(x, 10, -100, 100)

    assert poly.as_poly().degree() == 10
    assert all(-100 <= coeff <= 100 for coeff in poly.as_poly().coeffs()) is True

    poly = random_poly(x, 10, -100, 100, polys=True)

    assert poly.degree() == 10
    assert all(-100 <= coeff <= 100 for coeff in poly.coeffs()) is True

    poly = random_poly(x, 0, -10, 10, polys=True)

    assert poly.degree() == 0
    assert all(-10 <= c <= 10 for c in poly.all_coeffs())

    poly = random_poly(x, 1, -20, 20, polys=True)

    assert poly.degree() == 1
    assert all(-20 <= c <= 20 for c in poly.all_coeffs())

    poly = random_poly(x, 2, -30, 30, polys=True)

    assert poly.degree() == 2
    assert all(-30 <= c <= 30 for c in poly.all_coeffs())

    poly = random_poly(x, 3, -40, 40, polys=True)

    assert poly.degree() == 3
    assert all(-40 <= c <= 40 for c in poly.all_coeffs())

    poly = random_poly(x, 3, -400, 400, polys=True)

    assert poly.degree() == 3
    assert all(-400 <= c <= 400 for c in poly.all_coeffs())

    random.seed(11)
    assert random_poly(x, 10, -1, 1, polys=True).all_coeffs() == [1, 0, 1, 1,
                                                                  -1, 0, 0, -1,
                                                                  0, 0, 1]

    for _ in range(10):
        poly = random_poly(x, 3, -10, 10, percent=50, polys=True)
        assert poly.all_coeffs()[-1]
        assert len([c for c in poly.all_coeffs() if c == 0]) == 2


def test_interpolating_poly():
    x0, x1, x2, y0, y1, y2 = symbols('x:3, y:3')

    assert interpolating_poly(0, x) == 0
    assert interpolating_poly(1, x) == y0

    assert interpolating_poly(2, x) == \
        y0*(x - x1)/(x0 - x1) + y1*(x - x0)/(x1 - x0)

    assert interpolating_poly(3, x) == \
        y0*(x - x1)*(x - x2)/((x0 - x1)*(x0 - x2)) + \
        y1*(x - x0)*(x - x2)/((x1 - x0)*(x1 - x2)) + \
        y2*(x - x0)*(x - x1)/((x2 - x0)*(x2 - x1))


def test_fateman_poly_F_1():
    R, x, y = ring('x y', ZZ)

    f, g, h = R.fateman_poly_F_1()

    assert f == (1 + sum(R.gens))*(2 + sum(R.gens))
    assert g == (1 + sum(_**2 for _ in R.gens))*(-3*y*x**2 + y**2 - 1)
    assert h == 1

    R, x, y, *_ = ring('x y z t', ZZ)

    f, g, h = R.fateman_poly_F_1()

    assert f == (1 + sum(R.gens))*(2 + sum(R.gens))
    assert g == (1 + sum(_**2 for _ in R.gens))*(-3*y*x**2 + y**2 - 1)


def test_fateman_poly_F_2():
    R, x, _ = ring('x y', ZZ)

    f, g, h = R.fateman_poly_F_2()
    D = (1 + sum(R.gens))**2

    assert f == D*(-2 + x - sum(R.gens[1:]))**2
    assert g == D*(2 + sum(R.gens))**2
    assert h == D

    R, x, *_ = ring('x y z t', ZZ)

    f, g, h = R.fateman_poly_F_2()
    D = (1 + sum(R.gens))**2

    assert f == D*(-2 + x - sum(R.gens[1:]))**2
    assert g == D*(2 + sum(R.gens))**2
    assert h == D


def test_fateman_poly_F_3():
    R, x, _ = ring('x y', ZZ)

    f, g, h = R.fateman_poly_F_3()
    D = (1 + sum(_**R.ngens for _ in R.gens))**2

    assert f == D*(-2 + x**R.ngens - sum(_**R.ngens for _ in R.gens[1:]))**2
    assert g == D*(+2 + sum(_**R.ngens for _ in R.gens))**2
    assert h == D

    R, x, *_ = ring('x y z t', ZZ)

    f, g, h = R.fateman_poly_F_3()
    D = (1 + sum(_**R.ngens for _ in R.gens))**2

    assert f == D*(-2 + x**R.ngens - sum(_**R.ngens for _ in R.gens[1:]))**2
    assert g == D*(+2 + sum(_**R.ngens for _ in R.gens))**2
    assert h == D



import pytest

from diofant import Float, I, Matrix, Rational, Symbol, pi, sqrt
from diofant.geometry import Line, Point


__all__ = ()

x = Symbol('x', real=True)
y = Symbol('y', real=True)
z = Symbol('z', real=True)
t = Symbol('t', real=True)
k = Symbol('k', real=True)
x1 = Symbol('x1', real=True)
x2 = Symbol('x2', real=True)
x3 = Symbol('x3', real=True)
y1 = Symbol('y1', real=True)
y2 = Symbol('y2', real=True)
y3 = Symbol('y3', real=True)
z1 = Symbol('z1', real=True)
z2 = Symbol('z2', real=True)
z3 = Symbol('z3', real=True)


def feq(a, b):
    """Test if two floating point values are 'equal'."""
    t = Float('1.0E-10')
    return -t < a - b < t


def test_point():
    p1 = Point(x1, x2)
    p2 = Point(y1, y2)
    p3 = Point(0, 0)
    p4 = Point(1, 1)
    p5 = Point(0, 1)

    assert p1.origin == p3
    assert p1.ambient_dimension == 2
    assert p3.is_zero is True
    assert p1 in p1  # pylint: disable=comparison-with-itself
    assert p1 not in p2

    assert p1 in p1  # pylint: disable=comparison-with-itself
    assert p1 not in p2
    assert p2.y == y2
    assert (p3 + p4) == p4
    assert (p2 - p1) == Point(y1 - x1, y2 - x2)
    assert p4*5 == Point(5, 5)
    assert -p2 == Point(-y1, -y2)
    pytest.raises(ValueError, lambda: Point(3, I))
    pytest.raises(ValueError, lambda: Point(2*I, I))
    pytest.raises(ValueError, lambda: Point(3 + I, I))

    assert Point(34.05, sqrt(3)) == Point(Rational(681, 20), sqrt(3))
    assert Point.midpoint(p3, p4) == Point(0.5, 0.5)
    assert Point.midpoint(p1, p4) == Point(0.5 + x1/2, 0.5 + x2/2)
    assert Point.midpoint(p2, p2) == p2
    assert p2.midpoint(p2) == p2

    assert Point.distance(p3, p4) == sqrt(2)
    assert Point.distance(p1, p1) == 0
    assert Point.distance(p3, p2) == sqrt(p2.x**2 + p2.y**2)

    p1_1 = Point(x1, x1)
    p1_2 = Point(y2, y2)
    p1_3 = Point(x1 + 1, x1)
    assert Point.is_collinear() is False
    assert Point.is_collinear(p3)
    assert Point.is_collinear(p3, p4)
    assert Point.is_collinear(p3, p4, p1_1, p1_2)
    assert Point.is_collinear(p3, p4, p1_1, p1_3) is False
    assert Point.is_collinear(p3, p3, p4, p5) is False
    line = Line(Point(1, 0), slope=1)
    pytest.raises(TypeError, lambda: Point.is_collinear(line))
    pytest.raises(TypeError, lambda: p1_1.is_collinear(line))

    assert p3.intersection(Point(0, 0)) == [p3]
    assert p3.intersection(p4) == []

    x_pos = Symbol('x', real=True, positive=True)
    p2_1 = Point(x_pos, 0)
    p2_2 = Point(0, x_pos)
    p2_3 = Point(-x_pos, 0)
    p2_4 = Point(0, -x_pos)
    p2_5 = Point(x_pos, 5)
    assert Point.is_concyclic(p2_1)
    assert Point.is_concyclic(p2_1, p2_2)
    assert Point.is_concyclic(p2_1, p2_2, p2_3, p2_4)
    assert Point.is_concyclic(p2_1, p2_2, p2_3, p2_5) is False
    assert Point.is_concyclic(p4, p4 * 2, p4 * 3) is False
    pytest.raises(TypeError, lambda: Point.is_concyclic(p2_1, '123'))

    assert p4.scale(2, 3) == Point(2, 3)
    assert p3.scale(2, 3) == p3

    assert p4.rotate(pi, Point(0.5, 0.5)) == p3
    assert p1 + p2 == p1.midpoint(p2).scale(2, 2)
    assert p4 - (-p3) == p3.midpoint(p4).scale(2, 2)

    assert p4 * 5 == Point(5, 5)
    assert p4 / 5 == Point(0.2, 0.2)

    pytest.raises(ValueError, lambda: Point(0, 0) + 10)

    # Point differences should be simplified
    assert Point(x*(x - 1), y) - Point(x**2 - x, y + 1) == Point(0, -1)

    a, b = Rational(1, 2), Rational(1, 3)
    assert Point(a, b).evalf(2) == \
        Point(a.evalf(2), b.evalf(2))
    pytest.raises(ValueError, lambda: Point(1, 2) + 1)

    # test transformations
    p = Point(1, 0)
    assert p.rotate(pi/2) == Point(0, 1)
    assert p.rotate(pi/2, p) == p
    p = Point(1, 1)
    assert p.scale(2, 3) == Point(2, 3)
    assert p.translate(1, 2) == Point(2, 3)
    assert p.translate(1) == Point(2, 1)
    assert p.translate(y=1) == Point(1, 2)
    assert p.translate(*p.args) == Point(2, 2)

    # Check invalid input for transform
    pytest.raises(ValueError, lambda: p3.transform(p3))
    pytest.raises(ValueError, lambda: p.transform(Matrix([[1, 0], [0, 1]])))


def test_Point():
    p1 = Point(1, 5)
    p2 = Point(4, 2.5)
    p3 = Point(6.1, 3.5)
    assert p1.distance(p2) == sqrt(61)/2
    assert p2.distance(p3) == sqrt(541)/10
    assert p1.bounds == (1, 5, 1, 5)

    assert Point.is_concyclic() is False

    pytest.raises(ValueError, lambda: Point(1, 1, 1))
    pytest.raises(ValueError, lambda: Point(1, I))



"""Tests for Gosper's algorithm for hypergeometric summation.

Some tests were taken from :cite:`Petkovsek1997AeqB`.
"""

from diofant import Rational, binomial, factorial, pi, sin, sqrt
from diofant.abc import a, b, k, m, n, r, x
from diofant.concrete.gosper import gosper_sum


__all__ = ()


def test_gosper_sum():
    assert gosper_sum(1, (k, 0, n)) == 1 + n
    assert gosper_sum(k, (k, 0, n)) == n*(1 + n)/2
    assert gosper_sum(k**2, (k, 0, n)) == n*(1 + n)*(1 + 2*n)/6
    assert gosper_sum(k**3, (k, 0, n)) == n**2*(1 + n)**2/4

    assert gosper_sum(2**k, (k, 0, n)) == 2*2**n - 1

    assert gosper_sum(factorial(k), (k, 0, n)) is None
    assert gosper_sum(binomial(n, k), (k, 0, n)) is None

    assert gosper_sum(factorial(k)/k**2, (k, 0, n)) is None
    assert gosper_sum((k - 3)*factorial(k), (k, 0, n)) is None

    f = k*factorial(k)
    assert gosper_sum(f, k) == factorial(k)
    assert gosper_sum(f, (k, 0, n)) == n*factorial(n) + factorial(n) - 1

    f = (-1)**k*binomial(n, k)
    assert gosper_sum(f, (k, 0, n)) == 0
    assert gosper_sum(f, (k, 0, m)) == -(-1)**m*(m - n)*binomial(n, m)/n

    f = (4*k + 1)*factorial(k)/factorial(2*k + 1)
    assert gosper_sum(f, (k, 0, n)) == (2*factorial(2*n + 1) -
                                        factorial(n))/factorial(2*n + 1)
    assert gosper_sum(f,
                      (k, 3, n)) == (-60*factorial(n) +
                                     factorial(2*n + 1))/(60*factorial(2*n + 1))


def test_gosper_sum_indefinite():
    assert gosper_sum(k, k) == k*(k - 1)/2
    assert gosper_sum(k**2, k) == k*(k - 1)*(2*k - 1)/6

    assert gosper_sum(1/(k*(k + 1)), k) == -1/k
    assert gosper_sum(-(27*k**4 + 158*k**3 + 430*k**2 +
                        678*k + 445)*factorial(2*k + 3) /
                      (3*(3*k + 7)*factorial(3*k + 5)),
                      k) == (3*k + 5)*(k**2 + 2*k +
                                       5)*factorial(2*k + 3)/factorial(3*k + 5)


def test_gosper_sum_parametric():
    assert gosper_sum(binomial(Rational(1, 2),
                               m - k + 1)*binomial(Rational(1, 2), m + k),
                      (k, 1, n)) == (n*(1 + m - n)*(-1 + 2*m + 2*n) *
                                     binomial(Rational(1, 2), 1 + m - n) *
                                     binomial(Rational(1, 2),
                                              m + n)/(m*(1 + 2*m)))


def test_gosper_sum_algebraic():
    assert gosper_sum(n**2 + sqrt(2),
                      (n, 0, m)).equals((m + 1)*(2*m**2 + m + 6*sqrt(2))/6)


def test_gosper_sum_iterated():
    f1 = binomial(2*k, k)/4**k
    f2 = (1 + 2*n)*binomial(2*n, n)/4**n
    f3 = (1 + 2*n)*(3 + 2*n)*binomial(2*n, n)/(3*4**n)
    f4 = (1 + 2*n)*(3 + 2*n)*(5 + 2*n)*binomial(2*n, n)/(15*4**n)
    f5 = (1 + 2*n)*(3 + 2*n)*(5 + 2*n)*(7 + 2*n)*binomial(2*n, n)/(105*4**n)

    assert gosper_sum(f1, (k, 0, n)) == f2
    assert gosper_sum(f2, (n, 0, n)) == f3
    assert gosper_sum(f3, (n, 0, n)) == f4
    assert gosper_sum(f4, (n, 0, n)) == f5


def test_gosper_sum_AeqB_part1():
    # Ex. 5.7.1
    f1a = n**4
    f1b = n**3*2**n
    f1c = 1/(n**2 + sqrt(5)*n - 1)
    f1d = n**4*4**n/binomial(2*n, n)
    f1e = factorial(3*n)/(factorial(n)*factorial(n + 1)*factorial(n + 2)*27**n)
    f1f = binomial(2*n, n)**2/((n + 1)*4**(2*n))
    f1g = (4*n - 1)*binomial(2*n, n)**2/((2*n - 1)**2*4**(2*n))
    f1h = n*factorial(n - Rational(1, 2))**2/factorial(n + 1)**2

    g1a = m*(m + 1)*(2*m + 1)*(3*m**2 + 3*m - 1)/30
    g1b = 26 + 2**(m + 1)*(m**3 - 3*m**2 + 9*m - 13)
    g1c = (m + 1)*(m*(m**2 - 7*m + 3)*sqrt(5) - (
        3*m**3 - 7*m**2 + 19*m - 6))/(2*m**3*sqrt(5) + m**4 + 5*m**2 - 1)/6
    g1d = -Rational(2, 231) + 2*4**m*(m + 1)*(63*m**4 + 112*m**3 + 18*m**2 -
                                              22*m + 3)/(693*binomial(2*m, m))
    g1e = -Rational(9, 2) + (81*m**2 + 261*m + 200)*factorial(
        3*m + 2)/(40*27**m*factorial(m)*factorial(m + 1)*factorial(m + 2))
    g1f = (2*m + 1)**2*binomial(2*m, m)**2/(4**(2*m)*(m + 1))
    g1g = -binomial(2*m, m)**2/4**(2*m)
    g1h = (4*pi -
           (2*m + 1)**2*(3*m + 4)*factorial(m - Rational(1, 2))**2 /
           factorial(m + 1)**2)

    assert gosper_sum(f1a, (n, 0, m)).equals(g1a)
    assert gosper_sum(f1b, (n, 0, m)).equals(g1b)
    assert gosper_sum(f1c, (n, 0, m)).equals(g1c)
    assert gosper_sum(f1d, (n, 0, m)).equals(g1d)
    assert gosper_sum(f1e, (n, 0, m)).equals(g1e)
    assert gosper_sum(f1f, (n, 0, m)).equals(g1f)
    assert gosper_sum(f1g, (n, 0, m)).equals(g1g)
    assert gosper_sum(f1h, (n, 0, m)).equals(g1h)


def test_gosper_sum_AeqB_part2():
    # Ex. 5.7.2
    f2a = n**2*a**n
    f2b = (n - r/2)*binomial(r, n)
    f2c = factorial(n - 1)**2/(factorial(n - x)*factorial(n + x))
    f2d = n*(n + a + b)*a**n*b**n/(factorial(n + a)*factorial(n + b))

    g2a = -a*(a + 1)/(a - 1)**3 + a**(m + 1)*(a**2*m**2 - 2*a*m**2 + m**2 -
                                              2*a*m + 2*m + a + 1)/(a - 1)**3
    g2b = (m - r)*binomial(r, m)/2
    g2c = (factorial(m)**2/(x**2*factorial(-x + m)*factorial(x + m)) -
           sin(pi*x)/(pi*x**3))
    g2d = (1/(factorial(a - 1)*factorial(b - 1)) -
           a**(m + 1)*b**(m + 1)/(factorial(a + m)*factorial(b + m)))

    assert gosper_sum(f2a, (n, 0, m)).equals(g2a)
    assert gosper_sum(f2b, (n, 0, m)).equals(g2b)
    assert gosper_sum(f2c, (n, 1, m)).equals(g2c)
    assert gosper_sum(f2d, (n, 0, m)).equals(g2d)


def test_gosper_sum_AeqB_part3():
    # Ex. 5.7.3
    f3a = 1/n**4
    f3b = (6*n + 3)/(4*n**4 + 8*n**3 + 8*n**2 + 4*n + 3)
    f3c = 2**n*(n**2 - 2*n - 1)/(n**2*(n + 1)**2)
    f3d = n**2*4**n/((n + 1)*(n + 2))
    f3e = 2**n/(n + 1)
    f3f = 4*(n - 1)*(n**2 - 2*n - 1)/(n**2*(n + 1)**2*(n - 2)**2*(n - 3)**2)
    f3g = (n**4 - 14*n**2 - 24*n - 9)*2**n/(n**2*(n + 1)**2*(n + 2)**2 *
                                            (n + 3)**2)

    g3b = m*(m + 2)/(2*m**2 + 4*m + 3)
    g3c = 2**m/m**2 - 2
    g3d = Rational(2, 3) + 4**(m + 1)*(m - 1)/(m + 2)/3
    g3f = Rational(1, 16) - 1/((m - 2)**2*(m + 1)**2)  # sign corrected
    g3g = -Rational(2, 9) + 2**(m + 1)/((m + 1)**2*(m + 3)**2)

    assert gosper_sum(f3a, (n, 1, m)) is None
    assert gosper_sum(f3b, (n, 1, m)).equals(g3b)
    assert gosper_sum(f3c, (n, 1, m - 1)).equals(g3c)
    assert gosper_sum(f3d, (n, 1, m)).equals(g3d)
    assert gosper_sum(f3e, (n, 0, m - 1)) is None
    assert gosper_sum(f3f, (n, 4, m)).equals(g3f)
    assert gosper_sum(f3g, (n, 1, m)).equals(g3g)


def test_sympyissue_12018():
    # Ex. 5.7.11a
    f = binomial(n + 1, k)/2**(n + 1) - binomial(n, k)/2**n
    assert gosper_sum(f, (k, 0, n)).equals(-2**(-n - 1))



"""
Package for handling logical expressions.
"""

from .boolalg import (ITE, And, Equivalent, Implies, Nand, Nor, Not, Or, Xor,
                      false, simplify_logic, to_cnf, to_dnf, to_nnf, true)
from .inference import satisfiable


__all__ = ('ITE', 'And', 'Equivalent', 'Implies', 'Nand', 'Nor', 'Not', 'Or',
           'Xor', 'false', 'simplify_logic', 'to_cnf', 'to_dnf', 'to_nnf',
           'true', 'satisfiable')



import random

from ..config import query
from ..domains import ZZ
from .polyerrors import CoercionFailed, DomainError
from .rings import PolyElement, PolynomialRing
from .rootisolation import _FindRoot


class UnivarPolynomialRing(PolynomialRing, _FindRoot):
    """A class for representing univariate polynomial rings."""

    def __call__(self, element):
        if isinstance(element, list):
            try:
                return self.from_terms(element)
            except (TypeError, ValueError):
                return self.from_list(element)
        return super().__call__(element)

    def from_list(self, element):
        return self.from_dict({(i,): c for i, c in enumerate(element)})

    def _random(self, n, a, b, percent=None):
        domain = self.domain

        if percent is None:
            percent = 100//(b - a)
        percent = min(max(0, percent), 100)
        nz = ((n + 1)*percent)//100

        f = []
        while len(f) < n + 1:
            v = domain.convert(random.randint(a, b))
            if v:
                f.append(v)

        if nz:
            f[-nz:] = [domain.zero]*nz
            lt = f.pop(0)
            random.shuffle(f)
            f.insert(0, lt)

        return self.from_list(list(reversed(f)))

    def _gf_random(self, n, irreducible=False):
        domain = self.domain

        assert domain.is_FiniteField

        while True:
            f = [domain(random.randint(0, domain.order - 1))
                 for i in range(n)] + [domain.one]
            f = self.from_list(f)
            if not irreducible or f.is_irreducible:
                return f

    def dispersionset(self, p, q=None):
        r"""Compute the *dispersion set* of two polynomials.

        For two polynomials `f(x)` and `g(x)` with `\deg f > 0`
        and `\deg g > 0` the dispersion set `\operatorname{J}(f, g)` is defined as:

        .. math::
            \operatorname{J}(f, g)
            & := \{a \in \mathbb{N}_0 | \gcd(f(x), g(x+a)) \neq 1\} \\
            &  = \{a \in \mathbb{N}_0 | \deg \gcd(f(x), g(x+a)) \geq 1\}

        For a single polynomial one defines `\operatorname{J}(f) := \operatorname{J}(f, f)`.

        Examples
        ========

        Note that the definition of the dispersion is not symmetric:

        >>> R, x = ring('x', QQ)

        >>> fp = x**4 - 3*x**2 + 1
        >>> gp = fp.shift(-3)

        >>> R.dispersionset(fp, gp)
        {2, 3, 4}
        >>> R.dispersionset(gp, fp)
        set()

        Computing the dispersion also works over field extensions:

        >>> R, x = ring('x', QQ.algebraic_field(sqrt(5)))

        >>> fp = x**2 + sqrt(5)*x - 1
        >>> gp = x**2 + (2 + sqrt(5))*x + sqrt(5)

        >>> R.dispersionset(fp, gp)
        {2}
        >>> R.dispersionset(gp, fp)
        {1, 4}

        We can even perform the computations for polynomials
        having symbolic coefficients:

        >>> D, a = ring('a', QQ)
        >>> R, x = ring('x', D)

        >>> fp = 4*x**4 + (4*a + 8)*x**3 + (a**2 + 6*a + 4)*x**2 + (a**2 + 2*a)*x
        >>> R.dispersionset(fp)
        {0, 1}

        References
        ==========

        * :cite:`Man1994disp`
        * :cite:`Koepf98`
        * :cite:`Abramov71rat`
        * :cite:`Man1993indefsum`

        """
        # Check for valid input
        same = q is None
        if same:
            q = p

        if p.ring is not q.ring:
            raise ValueError('Polynomials must have the same ring')

        fdomain = self.domain.field

        # We define the dispersion of constant polynomials to be zero
        if p.degree() < 1 or q.degree() < 1:
            return {0}

        # Factor p and q over the rationals
        fp = p.factor_list()
        fq = q.factor_list() if not same else fp

        # Iterate over all pairs of factors
        J = set()
        for s, _ in fp[1]:
            for t, _ in fq[1]:
                m = s.degree()
                n = t.degree()
                if n != m:
                    continue
                an = s.LC
                bn = t.LC
                if an - bn:
                    continue
                # Note that the roles of `s` and `t` below are switched
                # w.r.t. the original paper. This is for consistency
                # with the description in the book of W. Koepf.
                anm1 = s[(m - 1,)]
                bnm1 = t[(n - 1,)]
                alpha = fdomain(anm1 - bnm1)/fdomain(n*bn)
                if alpha not in ZZ:
                    continue

                alpha = ZZ.convert(alpha)

                if alpha < 0 or alpha in J:
                    continue
                if n > 1 and s - t.shift(alpha):
                    continue
                J.add(alpha)

        return J


class UnivarPolyElement(PolyElement):
    """Element of univariate distributed polynomial ring."""

    def all_coeffs(self):
        if self:
            return [self[(i,)] for i in range(self.degree() + 1)]
        else:
            return [self.parent.domain.zero]

    def shift(self, a):
        return self.compose(0, self.ring.gens[0] + a)

    def half_gcdex(self, other):
        """
        Half extended Euclidean algorithm in `F[x]`.

        Returns ``(s, h)`` such that ``h = gcd(self, other)``
        and ``s*self = h (mod other)``.

        Examples
        ========

        >>> _, x = ring('x', QQ)

        >>> f = x**4 - 2*x**3 - 6*x**2 + 12*x + 15
        >>> g = x**3 + x**2 - 4*x - 4

        >>> f.half_gcdex(g)
        (-1/5*x + 3/5, x + 1)

        """
        ring = self.ring
        domain = ring.domain

        if not domain.is_Field:
            raise DomainError(f"can't compute half extended GCD over {domain}")

        a, b = ring.one, ring.zero
        f, g = self, other

        while g:
            q, r = divmod(f, g)
            f, g = g, r
            a, b = b, a - q*b

        a = a.quo_ground(f.LC)
        f = f.monic()

        return a, f

    @property
    def is_cyclotomic(self):
        return self.ring._cyclotomic_p(self)

    def _right_decompose(self, s):
        ring = self.ring
        domain = ring.domain
        x = ring.gens[0]

        n = self.degree()
        lc = self.LC

        f = self.copy()
        g = x**s

        r = n // s

        for i in range(1, s):
            coeff = domain.zero

            for j in range(i):
                if (n + j - i,) not in f:
                    continue

                assert (s - j,) in g

                fc, gc = f[(n + j - i,)], g[(s - j,)]
                coeff += (i - r*j)*fc*gc

            g[(s - i,)] = domain.quo(coeff, i*r*lc)

        g._strip_zero()

        return g

    def _left_decompose(self, h):
        ring = self.ring

        g, i = ring.zero, 0
        f = self.copy()

        while f:
            q, r = divmod(f, h)

            if r.degree() > 0:
                return
            else:
                g[(i,)] = r.LC
                f, i = q, i + 1

        g._strip_zero()

        return g

    def _decompose(self):
        df = self.degree()

        for s in range(2, df):
            if df % s != 0:
                continue

            h = self._right_decompose(s)
            g = self._left_decompose(h)

            if g is not None:
                return g, h

    def decompose(self):
        """
        Compute functional decomposition of ``f`` in ``K[x]``.

        Given a univariate polynomial ``f`` with coefficients in a field of
        characteristic zero, returns list ``[f_1, f_2, ..., f_n]``, where::

                  f = f_1 o f_2 o ... f_n = f_1(f_2(... f_n))

        and ``f_2, ..., f_n`` are monic and homogeneous polynomials of at
        least second degree.

        Unlike factorization, complete functional decompositions of
        polynomials are not unique, consider examples:

        1. ``f o g = f(x + b) o (g - b)``
        2. ``x**n o x**m = x**m o x**n``
        3. ``T_n o T_m = T_m o T_n``

        where ``T_n`` and ``T_m`` are Chebyshev polynomials.

        Examples
        ========

        >>> _, x = ring('x', ZZ)

        >>> (x**4 - 2*x**3 + x**2).decompose()
        [x**2, x**2 - x]

        References
        ==========

        * :cite:`Kozen1989decomposition`

        """
        F = []
        f = self.copy()

        while True:
            result = f._decompose()

            if result is not None:
                f, h = result
                F = [h] + F
            else:
                break

        return [f] + F

    def sturm(self):
        """
        Computes the Sturm sequence of ``f`` in ``F[x]``.

        Given a univariate, square-free polynomial ``f(x)`` returns the
        associated Sturm sequence (see e.g. :cite:`Davenport1988systems`)
        ``f_0(x), ..., f_n(x)`` defined by::

           f_0(x), f_1(x) = f(x), f'(x)
           f_n = -rem(f_{n-2}(x), f_{n-1}(x))

        Examples
        ========

        >>> _, x = ring('x', QQ)

        >>> (x**3 - 2*x**2 + x - 3).sturm()
        [x**3 - 2*x**2 + x - 3, 3*x**2 - 4*x + 1, 2/9*x + 25/9, -2079/4]

        """
        return self.ring._sturm(self)

    def __mul__(self, other):
        ring = self.ring
        try:
            other = ring.convert(other)
        except CoercionFailed:
            return NotImplemented
        if max(self.degree(), other.degree()) > query('KARATSUBA_CUTOFF'):
            return self._mul_karatsuba(other)
        return super().__mul__(other)

    def _mul_karatsuba(self, other):
        """
        Multiply dense polynomials in ``K[x]`` using Karatsuba's algorithm.

        References
        ==========

        * :cite:`Hoeven02`

        """
        ring = self.ring
        domain = ring.domain

        df = self.degree()
        dg = other.degree()

        n = max(df, dg) + 1

        n2 = n//2

        fl = self.slice(0, n2)
        gl = other.slice(0, n2)

        fh = self.slice(n2, n).quo_term(((n2,), domain.one))
        gh = other.slice(n2, n).quo_term(((n2,), domain.one))

        lo = fl*gl
        hi = fh*gh

        mid = (fl + fh)*(gl + gh)
        mid -= (lo + hi)

        return lo + mid.mul_monom((n2,)) + hi.mul_monom((2*n2,))



import functools
import operator
import random

from ..config import using
from ..core import Dummy
from ..domains.algebraicfield import AlgebraicElement
from ..integrals.heurisch import _symbols
from ..ntheory import nextprime
from .modulargcd import (_euclidean_algorithm, _gf_gcdex, _minpoly_from_dense,
                         _trunc)
from .polyerrors import NotInvertible, UnluckyLeadingCoefficient
from .polyutils import _sort_factors
from .rings import PolynomialRing
from .solvers import solve_lin_sys


# TODO
# ====

# -) efficiency of _factor can be improved for irreducible polynomials if the
#    univariate factorization is done before the LC is factored


def _alpha_to_z(f, ring):
    r"""
    Change the representation of a polynomial over
    `\mathbb Q(\alpha)` by replacing the algebraic element `\alpha` by a new
    variable `z`.

    Parameters
    ==========

    f : PolyElement
        polynomial in `\mathbb Q(\alpha)[x_0, \ldots, x_n]`
    ring : PolynomialRing
        the polynomial ring `\mathbb Q[x_0, \ldots, x_n, z]`

    Returns
    =======

    f_ : PolyElement
        polynomial in `\mathbb Q[x_0, \ldots, x_n, z]`

    """
    if isinstance(f, AlgebraicElement):
        ring = ring.drop(*ring.gens[:-1])
        f_ = ring(dict(f.rep))

    else:
        f_ = ring.zero

        for monom, coeff in f.items():
            coeff = coeff.rep.all_coeffs()
            n = len(coeff)

            for i in range(n):
                m = monom + (n-i-1,)
                if coeff[n - i - 1]:
                    f_[m] = coeff[n - i - 1]

    return f_


def _z_to_alpha(f, ring):
    r"""
    Change the representation of a polynomial in
    `\mathbb Q[x_0, \ldots, x_n, z]` by replacing the variable `z` by the
    algebraic element `\alpha` of the given ring
    `\mathbb Q(\alpha)[x_0, \ldots, x_n]`.

    Parameters
    ==========

    f : PolyElement
        polynomial in `\mathbb Q[x_0, \ldots, x_n, z]`
    ring : PolynomialRing
        the polynomial ring `\mathbb Q(\alpha)[x_0, \ldots, x_n]`

    Returns
    =======

    f_ : PolyElement
        polynomial in `\mathbb Q(\alpha)[x_0, \ldots, x_n]`

    """
    domain = ring.domain

    f_ = ring.zero
    for monom, coeff in f.items():
        m = monom[:-1]
        c = domain([0]*monom[-1] + [domain.domain(coeff)])

        if m not in f_:
            f_[m] = c
        else:
            f_[m] += c

    return f_


def _distinct_prime_divisors(S, domain):
    r"""
    Try to find pairwise coprime divisors of all elements of a given list
    `S` of integers.

    If this fails, ``None`` is returned.

    References
    ==========

    * :cite:`Javadi2009factor`

    """
    gcd = domain.gcd
    divisors = []

    for i, s in enumerate(S):
        divisors.append(s)

        for j in range(i):
            g = gcd(divisors[i], divisors[j])
            divisors[i] = divisors[i] // g
            divisors[j] = divisors[j] // g
            g1 = gcd(divisors[i], g)
            g2 = gcd(divisors[j], g)

            while g1 != 1:
                g1 = gcd(divisors[i], g1)
                divisors[i] = divisors[i] // g1

            while g2 != 1:
                g2 = gcd(divisors[j], g2)
                divisors[j] = divisors[j] // g2

            if divisors[i] == 1 or divisors[j] == 1:
                return

    return divisors


def _denominator(f):
    r"""
    Compute the denominator `\mathrm{den}(f)` of a polynomial `f` over
    `\mathbb Q`, i.e. the smallest integer such that `\mathrm{den}(f) f` is
    a polynomial over `\mathbb Z`.

    """
    ring = f.ring.domain.ring
    lcm = ring.lcm
    den = ring.one

    for coeff in f.values():
        den = lcm(den, coeff.denominator)

    return den


def _monic_associate(f, ring):
    r"""
    Compute the monic associate of a polynomial `f` over
    `\mathbb Q(\alpha)`, which is defined as

    .. math ::

        \mathrm{den}\left( \frac 1 {\mathrm{lc}(f)} f \right) \cdot \frac 1 {\mathrm{lc}(f)} f.

    The result is a polynomial in `\mathbb Z[x_0, \ldots, x_n, z]`.

    See also
    ========

    _denominator
    _alpha_to_z

    """
    qring = ring.clone(domain=ring.domain.field)
    f = _alpha_to_z(f.monic(), qring)
    f_ = f.clear_denoms()[1].set_ring(ring)

    return f_.primitive()[1]


def _leading_coeffs(f, U, gamma, lcfactors, A, D, denoms, divisors):
    r"""
    Compute the true leading coefficients in `x_0` of the irreducible
    factors of a polynomial `f`.

    If this fails, ``None`` is returned.

    Parameters
    ==========

    f : PolyElement
        squarefree polynomial in `Z[x_0, \ldots, x_n, z]`
    U : list of PolyElement objects
        monic univariate factors of `f(x_0, A)` in `\mathbb Q(\alpha)[x_0]`
    gamma : Integer
        integer content of `\mathrm{lc}_{x_0}(f)`
    lcfactors : list of (PolyElement, Integer) objects
        factorization of `\mathrm{lc}_{x_0}(f)` in
        `\mathbb Z[x_1, \ldots, x_n, z]`
    A : list of Integer objects
        the evaluation point `[a_1, \ldots, a_n]`
    D : Integer
        integral multiple of the defect of `\mathbb Q(\alpha)`
    denoms : list of Integer objects
        denominators of `\frac 1 {l(A)}` for `l` in ``lcfactors``
    divisors : list of Integer objects
        pairwise coprime divisors of all elements of ``denoms``

    Returns
    =======

    f : PolyElement
        possibly updated polynomial `f`
    lcs : list of PolyElement objects
        true leading coefficients of the irreducible factors of `f`
    U_ : list of PolyElement objects
        list of possibly updated monic associates of the univariate factors
        `U`

    References
    ==========

    * :cite:`Javadi2009factor`

    """
    ring = f.ring
    domain = ring.domain
    symbols = f.ring.symbols
    qring = ring.clone(symbols=(symbols[0], symbols[-1]), domain=domain.field)
    gcd = domain.gcd

    U = [_alpha_to_z(u, qring) for u, _ in U]
    denominators = [_denominator(u) for u in U]

    omega = D * gamma

    m = len(denoms)

    for i in range(m):
        pi = gcd(omega, divisors[i])
        divisors[i] //= pi
        if divisors[i] == 1:
            raise NotImplementedError

    e = []

    for dj in denominators:
        ej = []

        for i in range(m):
            eji = 0
            g1 = gcd(dj, divisors[i])

            while g1 != 1:
                eji += 1
                dj = dj // g1
                g1 = gcd(dj, g1)

            ej.append(eji)

        e.append(ej)

    n = len(denominators)
    if any(sum(e[j][i] for j in range(n)) != lcfactors[i][1] for i in range(m)):
        return

    lcring = ring.drop(0)
    lcs = []
    for j in range(n):
        lj = functools.reduce(operator.mul, [lcfactors[i][0]**e[j][i] for i in range(m)], lcring.one)
        lcs.append(lj)

    zring = qring.clone(domain=domain)

    for j in range(n):
        lj = lcs[j]
        dj = denominators[j]
        ljA = lj.eval(list(zip(lcring.gens, A)))

        lcs[j] = lj*dj
        U[j] = (U[j]*dj).set_ring(zring) * ljA.set_ring(zring)

        if omega == 1:
            f *= dj
        else:
            d = gcd(omega, dj)
            f *= (dj // d)

    if omega != 1:
        lcs[0] *= omega
        U[0] *= omega

    return f, lcs, U


def _test_evaluation_points(f, gamma, lcfactors, A, D):
    r"""
    Test if an evaluation point is suitable for _factor.

    If it is not, ``None`` is returned.

    Parameters
    ==========

    f : PolyElement
        squarefree polynomial in `\mathbb Z[x_0, \ldots, x_n, z]`
    gamma : Integer
        leading coefficient of `f` in `\mathbb Z`
    lcfactors : list of (PolyElement, Integer) objects
        factorization of `\mathrm{lc}_{x_0}(f)` in
        `\mathbb Z[x_1, \ldots, x_n, z]`
    A : list of Integer objects
        the evaluation point `[a_1, \ldots, a_n]`
    D : Integer
        integral multiple of the defect of `\mathbb Q(\alpha)`

    Returns
    =======

    fA : PolyElement
        `f` evaluated at `A`, i.e. `f(x_0, A)`
    denoms : list of Integer objects
        the denominators of `\frac 1 {l(A)}` for `l` in ``lcfactors``
    divisors : list of Integer objects
        pairwise coprime divisors of all elements of ``denoms``

    References
    ==========

    * :cite:`Javadi2009factor`

    See also
    ========

    _factor

    """
    ring = f.ring
    qring = ring.clone(domain=ring.domain.field)

    fA = f.eval(list(zip(ring.gens[1:-1], A)))

    if fA.degree() < f.degree():
        return

    if not fA.is_squarefree:
        return

    omega = gamma * D
    denoms = []
    for l, _ in lcfactors:
        lA = l.eval(list(zip(l.ring.gens, A)))  # in Q(alpha)
        denoms.append(_denominator(_alpha_to_z(lA**(-1), qring)))

    if any(denoms.count(denom) > 1 for denom in denoms):
        raise UnluckyLeadingCoefficient

    divisors = _distinct_prime_divisors(denoms, ring.domain)

    if divisors is None:
        return
    elif any(omega % d == 0 for d in divisors):
        return

    return fA, denoms, divisors


def _subs_ground(f, A):
    r"""
    Substitute variables in the coefficients of a polynomial `f` over a
    ``PolynomialRing``.

    """
    f_ = f.ring.zero

    for monom, coeff in f.items():
        if coeff.compose(A):
            f_[monom] = coeff.compose(A)

    return f_


def _padic_lift(f, pfactors, lcs, B, minpoly, p):
    r"""
    Lift the factorization of a polynomial over `\mathbb Z_p[z]/(\mu(z))` to
    a factorization over `\mathbb Z_{p^m}[z]/(\mu(z))`, where `p^m \geq B`.

    If this fails, ``None`` is returned.

    Parameters
    ==========

    f : PolyElement
        squarefree polynomial in `\mathbb Z[x_0, \ldots, x_n, z]`
    pfactors : list of PolyElement objects
        irreducible factors of `f` modulo `p`
    lcs : list of PolyElement objects
        true leading coefficients in `x_0` of the irreducible factors of `f`
    B : Integer
        heuristic numerical bound on the size of the largest integer
        coefficient in the irreducible factors of `f`
    minpoly : PolyElement
        minimal polynomial `\mu` of `\alpha` over `\mathbb Q`
    p : Integer
        prime number

    Returns
    =======

    H : list of PolyElement objects
        factorization of `f` modulo `p^m`, where `p^m \geq B`

    References
    ==========

    * :cite:`Javadi2009factor`

    """
    ring = f.ring
    domain = ring.domain

    x = ring.gens[0]
    tails = [g - g.eject(*ring.gens[1:]).LC.set_ring(ring)*x**g.degree() for g in pfactors]

    coeffs = []
    for i, g in enumerate(tails):
        coeffs += _symbols(f'c{i}', len(g))

    coeffring = PolynomialRing(domain, coeffs)
    ring_ = ring.clone(domain=coeffring)

    S = []
    k = 0
    for t in tails:
        s = ring_.zero
        r = len(t)
        for i, monom in zip(range(k, k + r), t):
            s[monom] = coeffring.gens[i]
        S.append(s)
        k += r

    m = minpoly.set_ring(ring_)
    f = f.set_ring(ring_)
    x = ring_.gens[0]
    H = [t.set_ring(ring_) + li.set_ring(ring_)*x**g.degree() for t, g, li in
         zip(tails, pfactors, lcs)]

    prod = functools.reduce(operator.mul, H)
    e = (f - prod) % m

    P = domain(p)
    while e and P < 2*B:
        poly = e // P

        for s, h in zip(S, H):
            poly -= (prod//h)*s

        poly = _trunc(poly, m, P)

        P_domain = domain.finite_ring(P)

        try:
            solution = solve_lin_sys([_.set_domain(P_domain)
                                      for _ in poly.values()],
                                     coeffring.clone(domain=P_domain))
        except NotInvertible:
            return

        if solution is None:
            return
        else:
            solution = {k.set_domain(domain): v.set_domain(domain).trunc_ground(P)
                        for k, v in solution.items()}
            assert len(solution) == coeffring.ngens

        subs = list(solution.items())

        H = [h + _subs_ground(s, subs)*P for h, s in zip(H, S)]
        P = P**2
        prod = functools.reduce(operator.mul, H)
        e = (f - prod) % m

    if e == 0:
        return [h.set_ring(ring) for h in H]
    else:
        return


def _div(f, g, minpoly, p):
    r"""
    Division with remainder for univariate polynomials over
    `\mathbb Z_p[z]/(\mu(z))`.

    """
    ring = f.ring
    domain = ring.domain

    rem = f
    deg = g.degree(0)
    lcinv, _, gcd = _gf_gcdex(g.eject(*ring.gens[1:]).LC, minpoly, p)

    if not gcd == 1:
        raise NotImplementedError

    quotient = ring.zero

    while True:
        degrem = rem.degree(0)
        if degrem < deg:
            break
        m = ring.from_terms([((degrem - deg, 0), domain.one)])
        quo = (lcinv * rem.eject(*ring.gens[1:]).LC).set_ring(ring)*m
        rem = _trunc(rem - g*quo, minpoly, p)
        quotient += quo

    return _trunc(quotient, minpoly, p), rem


def _extended_euclidean_algorithm(f, g, minpoly, p):
    r"""
    Extended Euclidean Algorithm for univariate polynomials over
    `\mathbb Z_p[z]/(\mu(z))`.

    Returns `s, t, h`, where `h` is the GCD of `f` and `g` and
    `sf + tg = h`.

    """
    ring = f.ring
    zero = ring.zero
    one = ring.one

    f = _trunc(f, minpoly, p)
    g = _trunc(g, minpoly, p)

    s0, s1 = zero, one
    t0, t1 = one, zero

    while g:
        result = _div(f, g, minpoly, p)
        if result is None:
            raise NotImplementedError
        quo, rem = result
        f, g = g, rem
        s0, s1 = s1 - quo*s0, s0
        t0, t1 = t1 - quo*t0, t0

    lcfinv = _gf_gcdex(f.eject(*ring.gens[1:]).LC, minpoly, p)[0].set_ring(ring)

    return (_trunc(s1 * lcfinv, minpoly, p), _trunc(t1 * lcfinv, minpoly, p),
            _trunc(f * lcfinv, minpoly, p))


def _diophantine_univariate(F, m, minpoly, p):
    r"""
    Solve univariate Diophantine equations of the form

    .. math ::

        \sum_{f \in F} \left( h_f(x) \cdot \prod_{g \in F \setminus \lbrace f \rbrace } g(x) \right) = x^m

    over `\mathbb Z_p[z]/(\mu(z))`.

    """
    ring = F[0].ring
    domain = ring.domain
    m = ring.from_terms([((m, 0), domain.one)])

    if len(F) == 2:
        f, g = F
        result = _extended_euclidean_algorithm(g, f, minpoly, p)
        if result is None:
            raise NotImplementedError
        s, t, _ = result

        s *= m
        t *= m

        q, s = _div(s, f, minpoly, p)

        t += q*g

        s = _trunc(s, minpoly, p)
        t = _trunc(t, minpoly, p)

        result = [s, t]
    else:
        G = [F[-1]]

        for f in reversed(F[1:-1]):
            G.insert(0, f * G[0])

        S, T = [], [ring.one]

        for f, g in zip(F, G):
            result = _diophantine([g, f], T[-1], [], 0, minpoly, p)
            if result is None:
                raise NotImplementedError
            t, s = result
            T.append(t)
            S.append(s)

        result, S = [], S + [T[-1]]

        for s, f in zip(S, F):
            r = _div(s*m, f, minpoly, p)[1]
            s = _trunc(r, minpoly, p)

            result.append(s)

    return result


def _diophantine(F, c, A, d, minpoly, p):
    r"""
    Solve multivariate Diophantine equations over `\mathbb Z_p[z]/(\mu(z))`.

    """
    ring = c.ring

    if not A:
        S = [ring.zero for _ in F]
        c = _trunc(c, minpoly, p)

        for (exp,), coeff in c.eject(1).items():
            T = _diophantine_univariate(F, exp, minpoly, p)
            if T is None:
                raise NotImplementedError

            for j, (s, t) in enumerate(zip(S, T)):
                S[j] = _trunc(s + t*coeff.set_ring(ring), minpoly, p)
    else:
        n = len(A)
        e = functools.reduce(operator.mul, F)

        a, A = A[-1], A[:-1]
        B, G = [], []

        for f in F:
            B.append(e//f)
            G.append(f.eval(n, a))

        C = c.eval(n, a)

        S = _diophantine(G, C, A, d, minpoly, p)
        if S is None:
            raise NotImplementedError
        S = [s.set_ring(ring) for s in S]

        for s, b in zip(S, B):
            c = c - s*b

        c = _trunc(c, minpoly, p)

        m = ring.gens[n] - a
        M = ring.one

        for k in range(d):
            if not c:
                break

            M = M * m
            C = c.diff(x=n, m=k + 1).eval(x=n, a=a)

            if C:
                C = C.quo_ground(ring.domain.factorial(k + 1))
                T = _diophantine(G, C, A, d, minpoly, p)
                if T is None:
                    raise NotImplementedError

                for i, t in enumerate(T):
                    T[i] = t.set_ring(ring) * M

                for i, (s, t) in enumerate(zip(S, T)):
                    S[i] = s + t

                for t, b in zip(T, B):
                    c = c - t * b

                c = _trunc(c, minpoly, p)
            else:
                raise NotImplementedError

        S = [_trunc(s, minpoly, p) for s in S]

    return S


def _hensel_lift(f, H, LC, A, minpoly, p):
    r"""
    Parallel Hensel lifting algorithm over `\mathbb  Z_p[z]/(\mu(z))`.

    Parameters
    ==========

    f : PolyElement
        squarefree polynomial in `\mathbb Z[x_0, \ldots, x_n, z]`
    H : list of PolyElement objects
        monic univariate factors of `f(x_0, A)` in
        `\mathbb Z[x_0, z]`
    LC : list of PolyElement objects
        true leading coefficients of the irreducible factors of `f`
    A : list of Integer objects
        the evaluation point `[a_1, \ldots, a_n]`
    p : Integer
        prime number

    Returns
    =======

    pfactors : list of PolyElement objects
        irreducible factors of `f` modulo `p`

    """
    ring = f.ring
    n = len(A)

    S = [f]
    H = list(H)

    for i, a in enumerate(reversed(A[1:])):
        s = S[0].eval(n - i, a)
        S.insert(0, _trunc(s, minpoly, p))

    d = max(f.degree(_) for _ in ring.gens[1:-1])

    for j, s, a in zip(range(1, n + 1), S, A):
        G = list(H)

        I, J = A[:j - 1], A[j:]

        Hring = f.ring
        for _ in range(j, n):
            Hring = Hring.drop(j + 1)

        x = Hring.gens[0]
        evalpoints = list(zip(LC[0].ring.gens[j:-1], J))

        for i, (h, lc) in enumerate(zip(H, LC)):
            if evalpoints:
                lc = lc.eval(evalpoints)
            lc = _trunc(lc, minpoly, p).set_ring(Hring)
            H[i] = h.set_ring(Hring) + (lc - h.eject(*h.ring.gens[1:]).LC.set_ring(Hring))*x**h.degree()

        m = Hring.gens[j] - a
        M = Hring.one

        c = _trunc(s - functools.reduce(operator.mul, H), minpoly, p)

        dj = s.degree(j)

        for k in range(dj):
            if not c:
                break

            M = M * m
            C = c.diff(x=j, m=k + 1).eval(x=j, a=a)

            if C:
                C = C.quo_ground(ring.domain.factorial(k + 1))  # coeff of (x_{j-1} - a_{j-1})^(k + 1) in c
                T = _diophantine(G, C, I, d, minpoly, p)
                if T is None:
                    raise NotImplementedError

                for i, (h, t) in enumerate(zip(H, T)):
                    H[i] = _trunc(h + t.set_ring(Hring)*M, minpoly, p)

                c = _trunc(s - functools.reduce(operator.mul, H), minpoly, p)

    prod = functools.reduce(operator.mul, H)
    if _trunc(prod, minpoly, p) == f.trunc_ground(p):
        return H


def _sqf_p(f, minpoly, p):
    r"""
    Return ``True`` if `f` is square-free in `\mathbb Z_p[z]/(\mu(z))[x]`.

    """
    ring = f.ring
    lcinv, *_ = _gf_gcdex(f.eject(*ring.gens[1:]).LC, minpoly, p)

    f = _trunc(f * lcinv.set_ring(ring), minpoly, p)

    if not f:
        return True
    else:
        return _euclidean_algorithm(f, _trunc(f.diff(0), minpoly, p), minpoly, p) == 1


def _test_prime(fA, D, minpoly, p, domain):
    r"""
    Test if a prime number is suitable for _factor.

    See also
    ========

    _factor

    """
    if fA.LC % p == 0 or minpoly.LC % p == 0:
        return False
    if not _sqf_p(fA, minpoly, p):
        return False
    if D % p == 0:
        return False

    return True


# squarefree f with cont_x0(f) = 1
def _factor(f, save):
    r"""
    Factor a multivariate polynomial `f`, which is squarefree and primitive
    in `x_0`, in `\mathbb Q(\alpha)[x_0, \ldots, x_n]`.

    References
    ==========

    * :cite:`Javadi2009factor`

    """
    ring = f.ring  # Q(alpha)[x_0, ..., x_{n-1}]
    lcring = ring.drop(0)
    uniring = ring.drop(*ring.gens[1:])
    ground = ring.domain.domain
    n = ring.ngens

    z = Dummy('z')
    qring = ring.clone(symbols=ring.symbols + (z,), domain=ground)
    lcqring = qring.drop(0)

    groundring = ground.ring
    zring = qring.clone(domain=groundring)
    lczring = zring.drop(0)

    minpoly = _minpoly_from_dense(ring.domain.mod, zring.drop(*zring.gens[:-1]))
    f_ = _monic_associate(f, zring)

    if save is True:
        D = minpoly.resultant(minpoly.diff(0))
    else:
        D = groundring.one

    # heuristic bound for p-adic lift
    B = (f_.max_norm() + 1)*D

    lc = f_.eject(*zring.gens[1:]).LC
    gamma, lcfactors = efactor(_z_to_alpha(lc, lcring))  # over QQ(alpha)[x_1, ..., x_n]

    gamma = ground.convert(gamma)
    D_ = gamma.denominator
    gamma_ = gamma.numerator
    lcfactors_ = []

    for l, exp in lcfactors:
        den, l_ = _alpha_to_z(l, lcqring).clear_denoms()  # l_ in QQ[x_1, ..., x_n, z], but coeffs in ZZ
        cont, l_ = l_.set_ring(lczring).primitive()
        D_ *= den**exp
        gamma_ *= cont**exp
        lcfactors_.append((l_, exp))

    f_ *= D_
    p = 2

    N = 0
    history = set()
    tries = 5  # how big should this be?

    while True:
        for _ in range(tries):
            A = tuple(random.randint(-N, N) for _ in range(n - 1))

            if A in history:
                continue
            history.add(A)

            try:
                result = _test_evaluation_points(f_, gamma_, lcfactors, A, D)
            except UnluckyLeadingCoefficient:
                # TODO: check interval
                C = [random.randint(1, 3*(N + 1)) for _ in range(n - 1)]
                gens = zring.gens
                x = gens[0]

                for i, ci in zip(range(1, n + 1), C):
                    xi = gens[i]
                    f_ = f_.compose(xi, x + xi*ci)

                lc, factors = _factor(_z_to_alpha(f_, ring), save)
                gens = factors[0].ring.gens
                x = gens[0]

                for i, ci in zip(range(1, n + 1), C):
                    xi = gens[i]
                    factors = [g.compose(xi, (xi - x).quo_ground(ci)) for g in factors]

                return lc, factors

            if result is None:
                continue
            fA, denoms, divisors = result

            with using(aa_factor_method='trager'):
                _, fAfactors = _z_to_alpha(fA, uniring).factor_list()
            if len(fAfactors) == 1:
                g = _z_to_alpha(f_, ring)
                return f.LC, [g.monic()]

            result = _leading_coeffs(f_, fAfactors, gamma_, lcfactors_, A, D, denoms, divisors)
            if result is None:
                continue
            f_, lcs, fAfactors_ = result

            prod = groundring.one
            for lc in lcs:
                prod *= lc.LC
            delta = (ground(prod, f_.LC)).numerator

            f_ *= delta

            while not _test_prime(fA, D, minpoly, p, zring.domain):
                p = nextprime(p)

            pfactors = _hensel_lift(f_, fAfactors_, lcs, A, minpoly, p)
            if pfactors is None:
                p = nextprime(p)
                f_ = f_.primitive()[1]
                continue

            factors = _padic_lift(f_, pfactors, lcs, B, minpoly, p)
            if factors is None:
                p = nextprime(p)
                f_ = f_.primitive()[1]
                B *= B
                continue

            return f.LC, [_z_to_alpha(g.primitive()[1], ring).monic() for g in factors]

        N += 1


# output of the form (lc, [(poly1, exp1), ...])
def efactor(f, save=True):
    r"""Factor a multivariate polynomial `f` in `\mathbb Q(\alpha)[x_0, \ldots, x_n]`.

    By default, an estimate of the defect of the algebraic field is included
    in all computations. If ``save`` is set to ``False``, the defect will be
    treated as one, thus computations are faster. However, if the defect of
    `\alpha` is larger than one, this may lead to wrong results.

    References
    ==========

    * :cite:`Javadi2009factor`

    """
    ring = f.ring

    assert ring.domain.is_AlgebraicField

    if f.is_ground:
        return f[1], []

    n = ring.ngens

    if n == 1:
        with using(aa_factor_method='trager'):
            return f.factor_list()
    else:
        cont, f = f.eject(*ring.gens[1:]).primitive()
        f = f.inject()
        if cont != 1:
            lccont, contfactors = efactor(cont)
            lc, factors = efactor(f)
            contfactors = [(g.set_ring(ring), exp) for g, exp in contfactors]
            return lccont * lc, _sort_factors(contfactors + factors)

        # this is only correct because the content in x_0 is already divided out
        lc, sqflist = f.sqf_list()
        factors = []
        for g, exp in sqflist:
            lcg, gfactors = _factor(g, save)
            lc *= lcg
            factors = factors + [(gi, exp) for gi in gfactors]

        return lc, _sort_factors(factors)



"""Sparse polynomial rings."""

from __future__ import annotations

import functools
import math
import operator

from ..config import query
from ..core import Expr, Integer, Symbol, cacheit
from ..core import symbols as _symbols
from ..core.compatibility import is_sequence
from ..core.sympify import CantSympify, sympify
from ..domains.compositedomain import CompositeDomain
from ..domains.domainelement import DomainElement
from ..domains.ring import CommutativeRing
from ..ntheory import multinomial_coefficients
from ..ntheory.modular import symmetric_residue
from .euclidtools import _GCD
from .factortools import _Factor
from .monomials import Monomial
from .orderings import lex
from .polyerrors import (CoercionFailed, ExactQuotientFailed, GeneratorsError,
                         GeneratorsNeeded, PolynomialDivisionFailed)
from .polyoptions import Domain as DomainOpt
from .polyoptions import Order as OrderOpt
from .specialpolys import _test_polys
from .sqfreetools import _SQF


def ring(symbols, domain, order=lex):
    """Construct a polynomial ring returning ``(ring, x_1, ..., x_n)``.

    Parameters
    ==========

    symbols : str, Symbol/Expr or sequence of str, Symbol/Expr (non-empty)
    domain : :class:`~diofant.domains.domain.Domain` or coercible
    order : :class:`~diofant.polys.polyoptions.Order` or coercible, optional, defaults to ``lex``

    Examples
    ========

    >>> R, x, y, z = ring('x y z', ZZ)
    >>> R
    ZZ[x,y,z]
    >>> x + y + z
    x + y + z

    """
    _ring = PolynomialRing(domain, symbols, order)
    return (_ring,) + _ring.gens


def _parse_symbols(symbols):
    if not symbols:
        raise GeneratorsNeeded("generators weren't specified")

    if isinstance(symbols, str):
        return _symbols(symbols, seq=True)
    elif isinstance(symbols, Expr):
        return symbols,
    elif is_sequence(symbols):
        if all(isinstance(s, str) for s in symbols):
            return _symbols(symbols)
        elif all(isinstance(s, Expr) for s in symbols):
            return symbols

    raise GeneratorsError('expected a string, Symbol or expression '
                          'or a non-empty sequence of strings, '
                          'Symbols or expressions')


class PolynomialRing(_GCD, CommutativeRing, CompositeDomain, _SQF, _Factor, _test_polys):
    """A class for representing multivariate polynomial rings."""

    is_PolynomialRing = True

    has_assoc_Ring = True

    def __new__(cls, domain, symbols, order=lex):
        from .univar import UnivarPolyElement, UnivarPolynomialRing

        symbols = tuple(_parse_symbols(symbols))
        ngens = len(symbols)
        domain = DomainOpt.preprocess(domain)
        order = OrderOpt.preprocess(order)

        new_cls = PolynomialRing if ngens > 1 else UnivarPolynomialRing

        key = new_cls.__name__, symbols, ngens, domain, order
        obj = _ring_cache.get(key)

        if obj is None:
            if isinstance(domain, CompositeDomain) and set(symbols) & set(domain.symbols):
                raise GeneratorsError("polynomial ring and it's ground domain share generators")

            obj = object.__new__(new_cls)
            obj._hash = hash(key)

            if new_cls == UnivarPolynomialRing:
                dtype = UnivarPolyElement
            else:
                dtype = PolyElement
            obj.dtype = type(dtype.__name__, (dtype,), {'ring': obj})

            obj.symbols = symbols
            obj.ngens = ngens
            obj.domain = domain
            obj.order = order

            obj.zero_monom = Monomial((0,)*ngens)

            gens = []
            one = domain.one
            expv = [0]*ngens
            for i in range(ngens):
                expv[i] = 1
                poly = obj.zero
                poly[expv] = one
                gens.append(poly)
                expv[i] = 0
            obj.gens = tuple(gens)

            obj.rep = str(domain) + '[' + ','.join(map(str, symbols)) + ']'

            for symbol, generator in zip(obj.symbols, obj.gens):
                if isinstance(symbol, Symbol):
                    name = symbol.name

                    if not hasattr(obj, name):
                        setattr(obj, name, generator)

            _ring_cache[key] = obj

        return obj

    def __getnewargs_ex__(self):
        return (self.domain, self.symbols), {'order': self.order}

    def leading_expv(self, f, order=None):
        order = self.order if order is None else OrderOpt.preprocess(order)
        return Monomial(max(f, key=order))

    @property
    def characteristic(self):
        return self.domain.characteristic

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self is other

    def clone(self, symbols=None, domain=None, order=None):
        return self.__class__(domain or self.domain, symbols or self.symbols, order or self.order)

    @property
    def zero(self):
        return self.dtype()

    @property
    def one(self):
        return self.ground_new(self.domain.one)

    def domain_new(self, element, orig_domain=None):
        return self.domain.convert(element, orig_domain)

    def ground_new(self, coeff):
        return self.term_new(self.zero_monom, coeff)

    def term_new(self, monom, coeff):
        poly = self.zero
        if coeff := self.domain.convert(coeff):
            poly[monom] = coeff
        return poly

    def __call__(self, element):
        if isinstance(element, PolyElement):
            if self == element.ring:
                return element
            elif isinstance(self.domain, PolynomialRing) and self.domain.ring == element.ring:
                return self.ground_new(element)
            else:
                raise NotImplementedError
        elif isinstance(element, str):
            raise NotImplementedError
        elif isinstance(element, dict):
            return self.from_dict(element)
        elif isinstance(element, list):
            return self.from_terms(element)
        elif isinstance(element, Expr):
            return self.convert(element)
        else:
            return self.ground_new(element)

    def from_dict(self, element):
        domain_new = self.domain.convert
        poly = self.zero

        for monom, coeff in element.items():
            if isinstance(monom, int):
                monom = monom,
            if coeff := domain_new(coeff):
                poly[monom] = coeff

        return poly

    def from_terms(self, element):
        return self.from_dict(dict(element))

    def from_expr(self, expr):
        expr = sympify(expr)
        mapping = dict(zip(self.symbols, self.gens))

        def _rebuild(expr):
            if (generator := mapping.get(expr)) is not None:
                return generator
            elif expr.is_Add:
                return functools.reduce(operator.add, map(_rebuild, expr.args))
            elif expr.is_Mul:
                return functools.reduce(operator.mul, map(_rebuild, expr.args))
            elif expr.is_Pow:
                c, a = expr.exp.as_coeff_Mul(rational=True)
                if c.is_Integer and c > 1:
                    return _rebuild(expr.base**a)**int(c)
            return self.ground_new(self.domain.convert(expr))

        try:
            return _rebuild(expr)
        except CoercionFailed as exc:
            raise ValueError('expected an expression convertible to a '
                             f'polynomial in {self}, got {expr}') from exc

    def index(self, gen):
        """Compute index of ``gen`` in ``self.gens``."""
        try:
            if isinstance(gen, int) and -self.ngens <= gen < self.ngens:
                return gen % self.ngens
            elif isinstance(gen, self.dtype):
                return self.gens.index(gen)
            elif isinstance(gen, str):
                return self.symbols.index(Symbol(gen))
            elif isinstance(gen, Expr):
                return self.symbols.index(gen)
        except ValueError:
            pass
        raise ValueError('expected a polynomial generator, an integer, '
                         f'a string, an expression or None, got {gen}')

    def drop(self, *gens):
        """Remove specified generators from this ring."""
        indices = set(map(self.index, gens))
        symbols = [s for i, s in enumerate(self.symbols) if i not in indices]

        if not symbols:
            return self.domain
        else:
            return self.clone(symbols=symbols)

    def to_ground(self):
        domain = self.domain
        if isinstance(domain, CompositeDomain) or domain.is_AlgebraicField:
            return self.clone(domain=domain.domain)
        else:
            raise ValueError(f'{domain} is not a composite or algebraic domain')

    @property
    def is_univariate(self):
        return len(self.gens) == 1

    @property
    def is_multivariate(self):
        return len(self.gens) > 1

    def eject(self, *gens):
        r"""
        Remove specified generators from the ring and inject them into
        its domain.

        """
        indices = set(map(self.index, gens))
        symbols = [s for i, s in enumerate(self.symbols) if i not in indices]
        gens = [gen for i, gen in enumerate(self.gens) if i not in indices]

        if not symbols:
            return self
        else:
            return self.clone(symbols=symbols, domain=self.drop(*gens))

    def to_expr(self, element):
        symbols = self.symbols
        domain = self.domain
        return functools.reduce(operator.add,
                                (domain.to_expr(v)*k.as_expr(*symbols)
                                 for k, v in element.items()),
                                Integer(0))

    def _from_PythonFiniteField(self, a, K0):
        if self.domain == K0:
            return self(a)
    _from_GMPYFiniteField = _from_PythonFiniteField
    _from_AlgebraicField = _from_PythonFiniteField
    _from_ExpressionDomain = _from_PythonFiniteField

    def _from_PythonIntegerRing(self, a, K0):
        return self(self.domain.convert(a, K0))
    _from_GMPYIntegerRing = _from_PythonIntegerRing
    _from_PythonRationalField = _from_PythonIntegerRing
    _from_GMPYRationalField = _from_PythonIntegerRing
    _from_RealField = _from_PythonIntegerRing
    _from_ComplexField = _from_PythonIntegerRing

    def _from_PolynomialRing(self, a, K0):
        try:
            return a.set_ring(self)
        except (CoercionFailed, GeneratorsError):
            return

    def _from_FractionField(self, a, K0):
        if self.domain == K0:
            return self.ground_new(a)

        (q,), r = a.numerator.div([a.denominator])

        if not r:
            return self.convert(q, K0.field.ring)

    @property
    def field(self):
        """Returns a field associated with ``self``."""
        return self.domain.frac_field(*self.symbols, order=self.order)

    def is_normal(self, a):
        return self.domain.is_normal(a.LC)

    def gcdex(self, a, b):
        """Extended GCD of ``a`` and ``b``."""
        return a.gcdex(b)

    def half_gcdex(self, a, b):
        """Half extended GCD of ``a`` and ``b``."""
        return a.half_gcdex(b)


_ring_cache: dict[tuple, PolynomialRing] = {}


class PolyElement(DomainElement, CantSympify, dict):
    """Element of multivariate distributed polynomial ring.

    Polynomial element is mutable, until the hash is computed, e.g.
    when the polynomial was added to the :class:`set`.

    If one is interested in preserving a polynomial, and one plans
    to use inplace operations, one can copy the polynomial first.

    Examples
    ========

    >>> _, x, y = ring('x y', ZZ)

    >>> p = (x + y)**2
    >>> p1 = p.copy()
    >>> p2 = p

    >>> p[(0, 0)] = 3
    >>> p1
    x**2 + 2*x*y + y**2
    >>> p2
    x**2 + 2*x*y + y**2 + 3

    >>> _ = hash(p)
    >>> p[(1, 1)] = 2
    Traceback (most recent call last):
    ...
    RuntimeError: ... Polynomial element ... can't be modified ...

    See Also
    ========

    PolynomialRing

    """

    @property
    def parent(self):
        return self.ring

    _hash = None

    def __hash__(self):
        _hash = self._hash
        if _hash is None:
            self._hash = _hash = hash((self.ring, frozenset(self.items())))
        return _hash

    def __reduce__(self):
        return self.parent.__call__, (dict(self),)

    def copy(self):
        """Return a shallow copy of self."""
        return self.__class__(self)

    def set_ring(self, new_ring):
        ring = self.ring
        symbols = ring.symbols
        new_symbols = new_ring.symbols

        if ring == new_ring:
            return self
        elif ring == new_ring.domain:
            return new_ring.ground_new(self)
        elif set(new_symbols).issuperset(symbols):
            coeffs = self.values()

            new_monoms = [[] for _ in range(len(self))]

            for gen in new_symbols:
                try:
                    j = symbols.index(gen)

                    for M, new_M in zip(self, new_monoms):
                        new_M.append(M[j])
                except ValueError:
                    for new_M in new_monoms:
                        new_M.append(0)

            terms = zip(map(Monomial, new_monoms), coeffs)

            return new_ring.from_terms(terms)
        else:
            raise CoercionFailed(f"Can't set element ring to {new_ring}")

    def set_domain(self, new_domain):
        if self.ring.domain == new_domain:
            return self
        else:
            new_ring = self.ring.clone(domain=new_domain)
            return self.set_ring(new_ring)

    def clear_denoms(self, convert=False):
        domain = self.ring.domain

        if not domain.is_Field:
            return domain.one, self

        if domain.has_assoc_Ring:
            ground_ring = domain.ring
        else:
            ground_ring = domain

        common = ground_ring.one
        lcm = ground_ring.lcm

        for coeff in self.values():
            common = lcm(common, coeff.denominator)

        f = self*domain.convert(common)

        if convert:
            f = f.set_domain(ground_ring)

        return common, f

    def _strip_zero(self):
        """Eliminate monomials with zero coefficient."""
        for k, v in list(self.items()):
            if not v:
                del self[k]

    def __setitem__(self, key, item):
        if self._hash is not None:
            raise RuntimeError(f"Polynomial element {self} can't be"
                               ' modified anymore.')
        if not isinstance(key, Monomial):
            key = Monomial(key)
        super().__setitem__(key, item)

    def __eq__(self, other):
        """Equality test for polynomials.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)
        >>> p1 = (x + y)**2 + (x - y)**2
        >>> p1 == 4*x*y
        False
        >>> p1 == 2*(x**2 + y**2)
        True

        """
        if not other:
            return not self
        elif isinstance(other, self.ring.dtype):
            return dict.__eq__(self, other)
        elif isinstance(other, self.ring.field.dtype):
            return other.__eq__(self)
        elif len(self) > 1:
            return False
        else:
            return self.get(self.ring.zero_monom) == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def drop(self, gen):
        ring = self.ring
        i = ring.index(gen)

        if ring.is_univariate:
            if self.is_ground:
                return self[1]
            else:
                raise ValueError(f"can't drop {gen}")
        else:
            symbols = list(ring.symbols)
            del symbols[i]
            ring = ring.clone(symbols=symbols)

            poly = ring.zero

            for k, v in self.items():
                if k[i] == 0:
                    K = list(k)
                    del K[i]
                    poly[K] = v
                else:
                    raise ValueError(f"can't drop {gen}")

            return poly

    def eject(self, *gens):
        ring = self.ring

        if not gens:
            return self

        if ring.is_univariate:
            raise ValueError("can't drop only generator to ground")

        indexes = [ring.index(gen) for gen in gens]
        ring = ring.eject(*indexes)

        poly = ring.zero
        gens = ring.domain.gens[0:len(indexes)]

        for monom, coeff in self.items():
            mon = tuple(monom[i] for i in range(self.ring.ngens) if i not in indexes)
            gc = functools.reduce(operator.mul, [x**n for x, n in zip(gens, (monom[i] for i in indexes))])
            if mon in poly:
                poly[mon] += gc*coeff
            else:
                poly[mon] = gc*coeff

        return poly

    def inject(self, front=False):
        ring = self.ring
        domain = ring.domain

        if not (isinstance(domain, CompositeDomain) or domain.is_AlgebraicField):
            return self

        new_ring = ring.to_ground()
        new_ring = new_ring.inject(*domain.symbols, front=front)

        poly = new_ring.zero

        for monom, coeff in self.items():
            coeff = coeff.to_dict()
            for cmonom, ccoeff in coeff.items():
                if front:
                    cmonom += monom
                else:
                    cmonom = monom + cmonom

                poly[cmonom] = ccoeff

        return poly

    def to_dict(self):
        return dict(self)

    def _str(self, printer, precedence, exp_pattern, mul_symbol):
        if not self:
            return printer._print(self.ring.domain.zero)
        prec_add = precedence['Add']
        prec_atom = precedence['Atom']
        ring = self.ring
        symbols = ring.symbols
        ngens = ring.ngens
        zm = ring.zero_monom
        order = ring.order
        sexpvs = []
        for expv, coeff in sorted(self.items(),
                                  key=lambda m: order(m[0]),
                                  reverse=True):
            normal = ring.domain.is_normal(coeff)
            sign = ' + ' if normal else ' - '
            sexpvs.append(sign)
            if expv == zm:
                scoeff = printer._print(coeff)
                if scoeff.startswith('-'):
                    scoeff = scoeff[1:]
            else:
                if not normal:
                    coeff = -coeff
                if coeff != 1:
                    scoeff = printer.parenthesize(coeff, prec_add)
                else:
                    scoeff = ''
            sexpv = []
            for i in range(ngens):
                exp = expv[i]
                if not exp:
                    continue
                symbol = printer.parenthesize(symbols[i], prec_atom-1)
                if exp != 1:
                    sexpv.append(exp_pattern % (symbol, exp))
                else:
                    sexpv.append(f'{symbol}')
            if scoeff:
                sexpv = [scoeff] + sexpv
            sexpvs.append(mul_symbol.join(sexpv))
        head = sexpvs.pop(0)
        if head == ' - ':
            sexpvs.insert(0, '-')
        return ''.join(sexpvs)

    @property
    def is_generator(self):
        return self.is_monomial and self.total_degree() == 1

    @property
    def is_ground(self):
        return not self or (len(self) == 1 and self.ring.zero_monom in self)

    @property
    def is_monomial(self):
        return not self or (len(self) == 1 and self.LC == 1)

    @property
    def is_term(self):
        return len(self) <= 1

    @property
    def is_linear(self):
        return all(sum(monom) <= 1 for monom in self)

    @property
    def is_quadratic(self):
        return all(sum(monom) <= 2 for monom in self)

    @property
    def is_irreducible(self):
        ring = self.ring
        domain = ring.domain

        if ring.is_univariate:
            if domain.is_FiniteField:
                method = query('GF_IRRED_METHOD')
                _irred_methods = {'ben-or': ring._gf_irreducible_p_ben_or,
                                  'rabin': ring._gf_irreducible_p_rabin}
                return _irred_methods[method](self)
            elif domain.is_IntegerRing:
                res = ring._zz_irreducible_p(self)
                if res is not None:
                    return res

        _, factors = self.factor_list()

        if not factors:
            return True
        elif len(factors) > 1:
            return False
        else:
            return factors[0][1] == 1

    @property
    def is_homogeneous(self):
        if not self:
            return True

        lm = self.LM
        tdeg = sum(lm)
        return all(sum(monom) == tdeg for monom in self if monom != lm)

    def __neg__(self):
        return self.__class__({monom: -self[monom] for monom in self})

    def __pos__(self):
        return self

    def __abs__(self):
        return self.__class__({monom: abs(self[monom]) for monom in self})

    def __add__(self, other):
        """Add two polynomials."""
        ring = self.ring
        try:
            other = ring.convert(other)
        except CoercionFailed:
            return NotImplemented
        result = self.copy()
        for t in other.items():
            result = result._iadd_term(t)
        return result

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        """Subtract polynomial other from self."""
        ring = self.ring
        try:
            other = ring.convert(other)
        except CoercionFailed:
            return NotImplemented
        result = self.copy()
        for k, v in other.items():
            result = result._iadd_term((k, -v))
        return result

    def __rsub__(self, other):
        """Substract self from other, with other convertible to the coefficient domain."""
        return (-self).__add__(other)

    def __mul__(self, other):
        """Multiply two polynomials."""
        ring = self.ring
        domain = ring.domain
        zero = ring.zero

        if not other:
            return zero
        elif isinstance(other, domain.dtype):
            result = ring.dtype({monom: self[monom]*other for monom in self})
            if not domain.is_Field and not domain.is_IntegerRing:
                result._strip_zero()
            return result

        try:
            other = ring.convert(other)
        except CoercionFailed:
            return NotImplemented

        if len(other) == 1:
            [(m, c)] = other.items()
            return self.__class__({monom*m: self[monom]*c for monom in self})

        result = zero
        for t in self.items():
            result = result._iadd_poly_term(other, t)
        return result

    def __rmul__(self, other):
        """Multiply other to self with other in the coefficient domain of self."""
        return self.__mul__(other)

    def __pow__(self, n, mod=None):
        """Raise polynomial to power `n`."""
        ring = self.ring

        if not isinstance(n, int) or n < 0:
            raise ValueError('exponent must be a nonnegative integer')

        if not n:
            return ring.one
        elif len(self) > 5 or mod:
            return self._pow_generic(n, mod)
        elif len(self) == 1:
            [(monom, coeff)] = self.items()
            p = ring.zero
            p[monom**n] = coeff**n
            return p
        elif n == 1:
            return self.copy()
        elif n == 2:
            return self._square()
        elif n == 3:
            return self*self._square()
        else:
            return self._pow_multinomial(n)

    def _pow_generic(self, n, mod=None):
        p = self.ring.one
        c = self

        while n:
            if n & 1:
                p *= c
                if mod:
                    p %= mod
                n -= 1

            c = c._square()
            if mod:
                c %= mod
            n //= 2

        return p

    def _pow_multinomial(self, n):
        multinomials = multinomial_coefficients(len(self), n).items()
        ring = self.ring
        zero_monom = ring.zero_monom
        terms = self.items()
        poly = ring.zero

        for multinomial, multinomial_coeff in multinomials:
            product_monom = zero_monom
            product_coeff = multinomial_coeff

            for exp, (monom, coeff) in zip(multinomial, terms):
                if exp:
                    product_monom *= monom**exp
                    product_coeff *= coeff**exp

            poly = poly._iadd_term((product_monom, product_coeff))

        return poly

    def _square(self):
        """Square of a polynomial."""
        ring = self.ring
        p = ring.zero
        get = p.get
        keys = list(self)
        zero = ring.domain.zero
        for i, k1 in enumerate(keys):
            pk = self[k1]
            for j in range(i):
                k2 = keys[j]
                exp = k1*k2
                p[exp] = get(exp, zero) + pk*self[k2]
        p += p
        get = p.get
        for k, v in self.items():
            k2 = k**2
            p[k2] = get(k2, zero) + v**2
        p._strip_zero()
        return p

    def __divmod__(self, other):
        ring = self.ring
        domain = ring.domain

        if not other:
            raise ZeroDivisionError('polynomial division')
        if isinstance(other, ring.dtype):
            (q,), r = self.div([other])
            return q, r
        elif isinstance(other, PolyElement):
            if isinstance(domain, PolynomialRing) and domain.ring == other.ring:
                pass
            else:
                return NotImplemented

        try:
            other = ring.domain_new(other)
        except CoercionFailed:
            return NotImplemented
        else:
            return self.quo_ground(other), self.trunc_ground(other)

    def __mod__(self, other):
        return divmod(self, other)[1]

    def __floordiv__(self, other):
        return divmod(self, other)[0]

    def __truediv__(self, other):
        ring = self.ring

        if not other:
            raise ZeroDivisionError('polynomial division')
        if isinstance(other, ring.domain.dtype):
            return self.quo_ground(other)

        try:
            other = ring.domain_new(other)
        except CoercionFailed:
            return NotImplemented
        else:
            return self.quo_ground(other)

    def div(self, fv):
        """Division algorithm for multivariate polynomials.

        Parameters
        ==========

        fv : sequence of PolyElement's
            List of divsors.

        Returns
        =======

        (qv, r) : tuple
            Where qv is the sequence of quotients and r is the remainder.

        Notes
        =====

        For multivariate polynomials the remainder is not uniquely
        determined, unless divisors form a Gröbner basis.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)
        >>> f = x**2*y
        >>> f1, f2 = x**2 - y, x*y - 1
        >>> f.div([f1, f2])
        ([y, 0], y**2)
        >>> f.div([f2, f1])
        ([x, 0], x)

        >>> g1, g2 = x - y**2, y**3 - 1
        >>> f.div([g1, g2])[1] == f.div([g2, g1])[1]
        True

        References
        ==========

        * :cite:`Cox2015ideals`, p. 64.

        """
        ring = self.ring
        order = ring.order
        if any(not f for f in fv):
            raise ZeroDivisionError('polynomial division')
        if any(f.ring != ring for f in fv):
            raise ValueError('self and f must have the same ring')
        if not self:
            return [ring.zero], ring.zero
        s = len(fv)
        qv = [ring.zero for i in range(s)]
        p = self.copy()
        r = ring.zero
        expvs = [fx.leading_expv() for fx in fv]
        while p:
            i = 0
            divoccurred = 0
            while i < s and divoccurred == 0:
                lt = p.leading_term()
                term = lt.quo_term((expvs[i], fv[i][expvs[i]]))
                if term:
                    expv1, c = term.LT
                    qv[i] = qv[i]._iadd_term((expv1, c))
                    p = p._iadd_poly_term(fv[i], (expv1, -c))
                    divoccurred = 1
                    if p and order(p.LM) >= order(lt.LM):
                        raise PolynomialDivisionFailed(self, fv[i], self.ring)
                else:
                    i += 1
            if not divoccurred:
                expv = p.leading_expv()
                r = r._iadd_term((expv, p[expv]))
                del p[expv]
        r._hash = None
        return qv, r

    def exquo(self, other):
        q, r = divmod(self, other)

        if not r:
            return q
        else:
            raise ExactQuotientFailed(self, other)

    def _iadd_term(self, term):
        """Add to self the term inplace.

        If self is a generator -- then just return the sum of the two.

        """
        ring = self.ring
        domain = ring.domain
        p1 = self
        if p1.is_generator:
            p1 = p1.copy()
        monom, coeff = term
        coeff += p1.get(monom, domain.zero)
        if coeff:
            p1[monom] = coeff
        elif monom in p1:
            del p1[monom]
        return p1

    def _iadd_poly_term(self, p2, term):
        """Add inplace to self the product of p2 and term.

        If self is a generator -- then just return the sum of the two.

        """
        p1 = self
        if p1.is_generator:
            p1 = p1.copy()
        monom, coeff = term
        for m, c in p2.items():
            m *= monom
            c *= coeff
            p1 = p1._iadd_term((m, c))
        return p1

    def degree(self, x=0):
        """
        The leading degree in ``x`` or the main variable.

        Note that the degree of 0 is negative floating-point infinity.

        """
        i = self.ring.index(x)
        return max((monom[i] for monom in self), default=-math.inf)

    def tail_degree(self, x=0):
        """
        The tail degree in ``x`` or the main variable.

        Note that the degree of 0 is negative floating-point infinity.

        """
        i = self.ring.index(x)
        return min((monom[i] for monom in self), default=-math.inf)

    def total_degree(self):
        """Returns the total degree."""
        return max((sum(m) for m in self), default=-math.inf)

    def leading_expv(self, order=None):
        """Leading monomial tuple according to the monomial ordering.

        Examples
        ========

        >>> _, x, y, z = ring('x y z', ZZ)
        >>> p = x**4 + x**3*y + x**2*z**2 + z**7
        >>> p.leading_expv()
        (4, 0, 0)

        """
        if self:
            return self.ring.leading_expv(self, order=order)

    def _get_coeff(self, expv):
        return self.get(expv, self.ring.domain.zero)

    def __getitem__(self, element):
        """
        Returns the coefficient that stands next to the given monomial.

        Parameters
        ==========

        element : PolyElement (with ``is_monomial = True``) or 1

        Examples
        ========

        >>> _, x, y, z = ring('x y z', ZZ)
        >>> f = 3*x**2*y - x*y*z + 7*z**3 + 23

        >>> f[x**2*y]
        3
        >>> f[x*y]
        0
        >>> f[1]
        23

        """
        ring = self.ring

        if isinstance(element, tuple):
            return self._get_coeff(element)
        elif element == 1:
            return self._get_coeff(ring.zero_monom)
        elif isinstance(element, ring.dtype) and element.is_monomial:
            monom, = element
            return self._get_coeff(monom)

        raise ValueError(f'expected a monomial, got {element}')

    @property
    def LC(self):
        return self._get_coeff(self.leading_expv())

    @property
    def LM(self):
        if (expv := self.leading_expv()) is None:
            return self.ring.zero_monom
        else:
            return expv

    @property
    def LT(self):
        if (expv := self.leading_expv()) is None:
            ring = self.ring
            return ring.zero_monom, ring.domain.zero
        else:
            return expv, self._get_coeff(expv)

    def leading_term(self, order=None):
        """Leading term as a polynomial element.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)
        >>> (3*x*y + y**2).leading_term()
        3*x*y

        """
        p = self.ring.zero
        if expv := self.leading_expv(order=order):
            p[expv] = self[expv]
        return p

    def content(self):
        """Returns GCD of polynomial's coefficients."""
        ring = self.ring
        domain = ring.domain
        gcd = domain.gcd

        cont = functools.reduce(gcd, self.values(), domain.zero)

        if not ring.is_normal(self):
            cont = -cont

        return cont

    def primitive(self):
        """Returns content and a primitive polynomial."""
        cont = self.content()
        prim = self.copy()
        if prim:
            prim = prim.quo_ground(cont)
        return cont, prim

    def monic(self):
        """Divides all coefficients by the leading coefficient."""
        if not self:
            return self
        else:
            return self.exquo_ground(self.LC)

    def mul_monom(self, m):
        return self.__class__({monom*m: self[monom] for monom in self})

    def quo_ground(self, x):
        domain = self.ring.domain

        if not x:
            raise ZeroDivisionError('polynomial division')
        if not self or x == 1:
            return self

        if domain.is_Field:
            quo = domain.quo
            p = self.__class__({monom: quo(self[monom], x) for monom in self})
        else:
            p = self.__class__({monom: self[monom]//x for monom in self})

        p._strip_zero()
        return p

    def exquo_ground(self, x):
        domain = self.ring.domain

        if not x:
            raise ZeroDivisionError('polynomial division')
        if not self or x == 1:
            return self

        p = self.__class__({monom: domain.exquo(self[monom], x) for monom in self})
        p._strip_zero()
        return p

    def quo_term(self, term):
        monom, coeff = term

        if not coeff:
            raise ZeroDivisionError('polynomial division')
        if not self:
            return self.ring.zero

        ring = self.ring
        domain = ring.domain
        p = ring.zero

        for tm, tc in self.items():
            if monom != self.ring.zero_monom:
                tm /= monom
            if any(_ < 0 for _ in tm):
                continue
            if domain.is_Field or not tc % coeff:
                p[tm] = domain.quo(tc, coeff)

        return p

    def trunc_ground(self, p):
        if self.ring.domain.is_IntegerRing:
            terms = {}

            for monom, coeff in self.items():
                coeff %= p
                terms[monom] = symmetric_residue(coeff, p)
        else:
            terms = {monom: self[monom] % p for monom in self}

        poly = self.__class__(terms)
        poly._strip_zero()
        return poly

    def _norm(self, norm_func):
        if not self:
            return self.ring.domain.zero
        else:
            return norm_func([abs(coeff) for coeff in self.values()])

    def max_norm(self):
        return self._norm(max)

    def l1_norm(self):
        return self._norm(sum)

    def gcd(self, other):
        return self.ring.gcd(self, other)

    def lcm(self, other):
        return self.ring.lcm(self, other)

    def cofactors(self, other):
        return self.ring.cofactors(self, other)

    def terms_gcd(self):
        ring = self.ring

        if not self:
            return (0,)*ring.ngens, self

        G = functools.reduce(Monomial.gcd, self)

        if all(g == 0 for g in G):
            return G, self

        f = ring.zero

        for monom, coeff in self.items():
            f[monom/G] = coeff

        return G, f

    def cancel(self, g, include=True):
        """
        Cancel common factors in a rational function ``f/g``.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)

        >>> (2*x**2 - 2).cancel(x**2 - 2*x + 1)
        (2*x + 2, x - 1)

        """
        f = self
        ring = f.ring
        domain = ring.domain

        if not domain.is_Field or not domain.has_assoc_Ring:
            _, p, q = f.cofactors(g)
            cp, cq = domain.one, domain.one
        else:
            new_ring = ring.clone(domain=domain.ring)

            cq, f = f.clear_denoms(convert=True)
            cp, g = g.clear_denoms(convert=True)

            _, p, q = f.cofactors(g)
            _, cp, cq = new_ring.domain.cofactors(cp, cq)

            p = p.set_ring(ring)
            q = q.set_ring(ring)

        p_neg = not ring.is_normal(p)
        q_neg = not ring.is_normal(q)

        if p_neg and q_neg:
            p, q = -p, -q
        elif p_neg:
            cp, p = -cp, -p
        elif q_neg:
            cp, q = -cp, -q

        if not include:
            return cp, cq, p, q

        p *= domain(cp)
        q *= domain(cq)

        return p, q

    def diff(self, x=0, m=1):
        """Computes partial derivative in ``x``.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)
        >>> p = x + x**2*y**3
        >>> p.diff(x)
        2*x*y**3 + 1

        """
        ring = self.ring
        i = ring.index(x)
        x, = ring.gens[i]
        x = x**m
        g = ring.zero if m else self.compose(ring.gens[i], ring.zero)
        for expv, coeff in self.items():
            if expv[i]:
                e = expv/x
                for j in range(expv[i], expv[i] - m, -1):
                    coeff *= j
                g[e] = coeff
        g._strip_zero()
        return g

    def integrate(self, x=0, m=1):
        """Computes indefinite integral in ``x``."""
        ring = self.ring
        i = ring.index(x)
        x, = ring.gens[i]
        x = x**m
        g = ring.zero
        for expv, coeff in self.items():
            e = expv*x
            for j in range(expv[i] + m, expv[i], -1):
                coeff /= j
            g[e] = coeff
        g._strip_zero()
        return g

    def __call__(self, *values):
        if 0 < len(values) <= self.ring.ngens:
            return self.eval(list(zip(self.ring.gens, values)))
        else:
            raise ValueError(f'expected at least 1 and at most {self.ring.ngens} values, got {len(values)}')

    def eval(self, x=0, a=0):
        if isinstance(x, list) and not a:
            (X, a), x = x[0], x[1:]
            f = self.eval(X, a)

            if x:
                return f.eval([(Y.drop(X), a) for (Y, a) in x])
            else:
                return f

        return self.compose(x, a).drop(x)

    def compose(self, x, a=None):
        """Computes the functional composition."""
        ring = self.ring
        poly = ring.zero

        if a is not None:
            replacements = [(x, a)]
        else:
            if isinstance(x, list):
                replacements = list(x)
            elif isinstance(x, dict):
                replacements = list(x.items())
            else:
                raise ValueError('expected a generator, value pair a '
                                 'sequence of such pairs')

        replacements = [(ring.index(x), ring(g)) for x, g in replacements]
        replacements.sort(key=lambda k: k[0])

        if ring.is_univariate:
            [(i, g)] = replacements
            acc, d = ring.one, 0
            for monom, coeff in sorted(self.items(), key=lambda x: x[0]):
                n = monom[i]
                acc *= g**(n - d)
                d = n
                poly += acc*coeff
            return poly

        for monom, coeff in self.items():
            monom = list(monom)
            subpoly = ring.one

            for i, g in replacements:
                n, monom[i] = monom[i], 0
                subpoly *= g**n

            monom = Monomial(monom)
            subpoly *= ring.from_terms([(monom, coeff)])
            poly += subpoly

        return poly

    def discriminant(self):
        """Computes discriminant of a polynomial."""
        ring = self.ring

        if (d := self.degree()) <= 0:
            return ring.zero.drop(0)
        else:
            s = (-1)**((d*(d - 1)) // 2)
            c = self.eject(*ring.gens[1:]).LC

            return self.resultant(self.diff()) // (c*s)

    def slice(self, m, n, x=0):
        ring = self.ring
        poly = ring.zero
        j = ring.index(x)

        for monom, coeff in self.items():
            if not n > monom[j] >= m:
                if ring.is_univariate:
                    continue
                monom = monom[:j] + (0,) + monom[j + 1:]

            if monom in poly:
                poly[monom] += coeff
            else:
                poly[monom] = coeff

        return poly

    def prem(self, other):
        """Polynomial pseudo-remainder.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)

        >>> (x**2 + x*y).prem(2*x + 2)
        -4*y + 4

        References
        ==========

        * :cite:`Knuth1985seminumerical`, p. 407.

        """
        ring = self.ring

        if not isinstance(other, ring.dtype):
            other = ring.convert(other)

        f, g = self, other

        if ring.is_multivariate:
            f, g = map(operator.methodcaller('eject', *ring.gens[1:]), (f, g))
            r = f.prem(g)
            return r.inject()

        df = f.degree()
        dg = g.degree()

        if dg < 0:
            raise ZeroDivisionError('polynomial division')

        r, dr = f, df

        if dr < dg:
            return r

        x = ring.gens[0]
        n = df - dg + 1
        lc_g = g.LC

        while True:
            lc_r = r.LC
            n -= 1

            r *= lc_g
            r -= g*x**(dr - dg)*lc_r
            dr = r.degree()

            if dr < dg:
                break

        r *= lc_g**n

        return r

    @cacheit
    def resultant(self, other, includePRS=False):
        """
        Computes resultant of two polynomials in `K[X]`.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)

        >>> f = 3*x**2*y - y**3 - 4
        >>> g = x**2 + x*y**3 - 9

        >>> f.resultant(g)
        -3*y**10 - 12*y**7 + y**6 - 54*y**4 + 8*y**3 + 729*y**2 - 216*y + 16

        """
        ring = self.ring
        domain = ring.domain

        if (not includePRS and query('USE_COLLINS_RESULTANT') and
                (domain.is_IntegerRing or domain.is_RationalField)):
            return ring._collins_resultant(self, other)

        res = ring._primitive_prs(self, other)

        return res if includePRS else res[0]

    def subresultants(self, other):
        """
        Computes subresultant PRS of two polynomials in `K[X]`.

        Examples
        ========

        >>> _, x, y = ring('x y', ZZ)

        >>> f = 3*x**2*y - y**3 - 4
        >>> g = x**2 + x*y**3 - 9

        >>> a = 3*x*y**4 + y**3 - 27*y + 4
        >>> b = -3*y**10 - 12*y**7 + y**6 - 54*y**4 + 8*y**3 + 729*y**2 - 216*y + 16

        >>> f.subresultants(g) == [f, g, a, b]
        True

        """
        return self.resultant(other, includePRS=True)[1]

    def gcdex(self, other):
        """
        Extended Euclidean algorithm in `F[x]`.

        Returns ``(s, t, h)`` such that ``h = gcd(self, other)`` and
        ``s*self + t*other = h``.

        Examples
        ========

        >>> _, x = ring('x', QQ)

        >>> f = x**4 - 2*x**3 - 6*x**2 + 12*x + 15
        >>> g = x**3 + x**2 - 4*x - 4

        >>> f.gcdex(g)
        (-1/5*x + 3/5, 1/5*x**2 - 6/5*x + 2, x + 1)

        """
        s, h = self.half_gcdex(other)
        t = h - self*s
        t //= other
        return s, t, h

    def sqf_list(self):
        return self.ring.sqf_list(self)

    def sqf_part(self):
        return self.ring.sqf_part(self)

    @property
    def is_squarefree(self):
        return self.ring.is_squarefree(self)

    def sqf_norm(self):
        return self.ring.sqf_norm(self)

    def factor_list(self):
        return self.ring.factor_list(self)



"""Solvers of systems of polynomial equations."""

import collections

from ..core import expand_mul
from ..domains import EX
from ..matrices import Matrix
from ..polys import (ComputationFailed, PolificationFailed, groebner,
                     parallel_poly_from_expr)
from ..polys.solvers import solve_lin_sys
from ..simplify.simplify import simplify
from ..utilities import default_sort_key, numbered_symbols
from .utils import checksol


__all__ = ('solve_linear_system', 'solve_poly_system',
           'solve_surd_system')


def solve_linear_system(system, *symbols, **flags):
    r"""Solve system of linear equations.

    Both under- and overdetermined systems are supported. The possible
    number of solutions is zero, one or infinite.

    Parameters
    ==========

    system : Matrix
        Nx(M+1) matrix, which means it has to be in augmented
        form.  This matrix will not be modified.
    \*symbols : list
        List of M Symbol's

    Returns
    =======

    solution: dict or None
        Respectively, this procedure will return None or
        a dictionary with solutions.  In the case of underdetermined
        systems, all arbitrary parameters are skipped.  This may
        cause a situation in which an empty dictionary is returned.
        In that case, all symbols can be assigned arbitrary values.

    Examples
    ========

    Solve the following system::

           x + 4 y ==  2
        -2 x +   y == 14

    >>> system = Matrix(((1, 4, 2), (-2, 1, 14)))
    >>> solve_linear_system(system, x, y)
    {x: -6, y: 2}

    A degenerate system returns an empty dictionary.

    >>> system = Matrix(((0, 0, 0), (0, 0, 0)))
    >>> solve_linear_system(system, x, y)
    {}

    See Also
    ========

    diofant.matrices.matrices.MatrixBase.rref

    """
    eqs = system*Matrix(symbols + (-1,))
    polys, _ = parallel_poly_from_expr([expand_mul(e) for e in eqs],
                                       *symbols, field=True)
    domain = polys[0].rep.ring
    polys = [_.rep for _ in polys]

    res = solve_lin_sys(polys, domain)
    if res is None:
        return

    for k in list(res):
        s = domain.symbols[domain.index(k)]
        res[s] = domain.to_expr(res[k])
        del res[k]
        if flags.get('simplify', True):
            res[s] = simplify(res[s])

    return res


def solve_poly_system(eqs, *gens, **args):
    """
    Solve a system of polynomial equations.

    Polynomial system may have finite number of solutions or
    infinitely many (positive-dimensional systems).

    References
    ==========

    * :cite:`Cox2015ideals`, p. 98

    Examples
    ========

    >>> solve_poly_system([x*y - 2*y, 2*y**2 - x**2], x, y)
    [{x: 0, y: 0}, {x: 2, y: -sqrt(2)}, {x: 2, y: sqrt(2)}]

    >>> solve_poly_system([x*y], x, y)
    [{x: 0}, {y: 0}]

    """
    try:
        args['extension'] = False
        polys, opt = parallel_poly_from_expr(eqs, *gens, **args)
        polys = [p.to_exact() for p in polys]
    except PolificationFailed as exc:
        raise ComputationFailed('solve_poly_system', len(eqs), exc) from exc

    def _solve_reduced_system(system, gens):
        """Recursively solves reduced polynomial systems."""
        basis = groebner(system, *gens, polys=True, extension=False)
        dim = basis.dimension
        solutions = []

        if dim is None:
            return []

        elif dim > 0:
            max_iset = max(basis.independent_sets, key=len)
            new_gens = [g for g in gens if g not in max_iset]

            # After removing variables from the maximal set of independent
            # variables for the given ideal - the new ideal is of dimension
            # zero with the independent variables as parameters in the
            # coefficient domain.
            solutions.extend(_solve_reduced_system(system, new_gens))

            # Now we should examine cases when leading coefficient of
            # some polynomial in the system is zero.
            for p in basis.polys:
                lc = p.as_poly(*new_gens).LC(order=basis.order)
                for special in _solve_reduced_system(system + [lc], gens):
                    # This heuristics wipe out some redundant special
                    # solutions, which already there in solutions after
                    # solving the system with new set of generators.
                    if all(any((_.subs(s) - _).subs(special).simplify()
                               for _ in gens) for s in solutions):
                        solutions.insert(0, special)

        else:
            # By the elimination property, the last polynomial should
            # be univariate in the last variable.
            f = basis[-1]
            gen = gens[-1]

            zeros = {k.doit() for k in f.exclude().all_roots()}

            if len(basis) == 1:
                return [{gen: zero} for zero in zeros]

            new_basis = [b.set_domain(EX) for b in basis[:-1]]

            # Now substitute zeros for the last variable and
            # solve recursively new obtained zero-dimensional systems.
            for zero in zeros:
                new_system = []
                new_gens = gens[:-1]

                for b in new_basis:
                    eq = b.eval(gen, zero)

                    if not eq.is_zero:
                        new_system.append(eq)

                for solution in _solve_reduced_system(new_system, new_gens):
                    solution[gen] = zero
                    solutions.append(solution)

        return solutions

    result = _solve_reduced_system(polys, opt.gens)

    if not opt.domain.is_Exact:
        result = [{k: r[k].evalf(opt.domain.dps) for k in r} for r in result]

    return sorted(result, key=default_sort_key)


def solve_surd_system(eqs, *gens, **args):
    """
    Solve a system of algebraic equations.

    Examples
    ========

    >>> solve_surd_system([x + sqrt(x + 1) - 2])
    [{x: -sqrt(13)/2 + 5/2}]

    """
    eqs = list(eqs)

    if not gens:
        gens = set().union(*[_.free_symbols for _ in eqs])
        gens = sorted(gens, key=default_sort_key)
    else:
        gens = list(gens)

    aux = numbered_symbols('a')
    neqs = len(eqs)
    orig_eqs = eqs[:]
    ngens = len(gens)
    bases = collections.defaultdict(dict)

    def q_surd(e):
        return e.is_Pow and e.exp.is_Rational and not e.exp.is_Integer

    def tr_surd(e):
        n, d = e.exp.as_numer_denom()
        for v2, d2 in sorted(bases.get(e.base, {}).items(),
                             key=lambda _: -_[1]):
            if not d2 % d:
                return v2**(d2 // d)
        v = next(aux)
        bases[e.base][v] = d
        gens.append(v)
        eqs.append(v**d - e.base)
        return v**n

    for i in range(neqs):
        eqs[i] = eqs[i].replace(q_surd, tr_surd)

    denoms = []
    for i, e in enumerate(eqs):
        eqs[i], d = e.as_numer_denom()
        if not d.is_constant(*gens):
            denoms.insert(0, d)

    weaksols = solve_poly_system(eqs, *gens, **args)

    for i in range(len(weaksols) - 1, -1, -1):
        if any(checksol(_, weaksols[i], warn=True) for _ in denoms):
            del weaksols[i]
        elif any(checksol(_, weaksols[i], warn=True) is False for _ in orig_eqs):
            del weaksols[i]
        else:
            for g in gens[ngens:]:
                del weaksols[i][g]

    return weaksols
